from contextlib import contextmanager
import psycopg2
import psycopg2.extras
from parsons.etl.table import Table
from parsons.utilities import files
import pickle
import petl
import logging
from parsons.databases.postgres.postgres_create_statement import PostgresCreateStatement

# Max number of rows that we query at a time, so we can avoid loading huge
# data sets into memory.
# 100k rows per batch at ~1k bytes each = ~100MB per batch.
QUERY_BATCH_SIZE = 100000

logger = logging.getLogger(__name__)


class PostgresCore(PostgresCreateStatement):

    @contextmanager
    def connection(self):
        """
        Generate a Postgres connection.
        The connection is set up as a python "context manager", so it will be closed
        automatically (and all queries committed) when the connection goes out of scope.

        When using the connection, make sure to put it in a ``with`` block (necessary for
        any context manager):
        ``with pg.connection() as conn:``

        `Returns:`
            Psycopg2 `connection` object
        """

        # Create a psycopg2 connection and cursor
        conn = psycopg2.connect(user=self.username, password=self.password,
                                host=self.host, dbname=self.db, port=self.port,
                                connect_timeout=self.timeout)

        try:
            yield conn
        except psycopg2.Error:
            conn.rollback()
            raise
        else:
            conn.commit()
        finally:
            conn.close()

    @contextmanager
    def cursor(self, connection):
        cur = connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        try:
            yield cur
        finally:
            cur.close()

    def query(self, sql, parameters=None):
        """
        Execute a query against the database. Will return ``None``if the query returns zero rows.

        To include python variables in your query, it is recommended to pass them as parameters,
        following the `psycopg style <http://initd.org/psycopg/docs/usage.html#passing-parameters-to-sql-queries>`_.
        Using the ``parameters`` argument ensures that values are escaped properly, and avoids SQL
        injection attacks.

        **Parameter Examples**

        .. code-block:: python

            # Note that the name contains a quote, which could break your query if not escaped
            # properly.
            name = "Beatrice O'Brady"
            sql = "SELECT * FROM my_table WHERE name = %s"
            rs.query(sql, parameters=[name])

        .. code-block:: python

            names = ["Allen Smith", "Beatrice O'Brady", "Cathy Thompson"]
            placeholders = ', '.join('%s' for item in names)
            sql = f"SELECT * FROM my_table WHERE name IN ({placeholders})"
            rs.query(sql, parameters=names)

        `Args:`
            sql: str
                A valid SQL statement
            parameters: list
                A list of python variables to be converted into SQL values in your query

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.

        """  # noqa: E501

        with self.connection() as connection:
            return self.query_with_connection(sql, connection, parameters=parameters)

    def query_with_connection(self, sql, connection, parameters=None, commit=True):
        """
        Execute a query against the database, with an existing connection. Useful for batching
        queries together. Will return ``None`` if the query returns zero rows.

        `Args:`
            sql: str
                A valid SQL statement
            connection: obj
                A connection object obtained from ``redshift.connection()``
            parameters: list
                A list of python variables to be converted into SQL values in your query
            commit: boolean
                Whether to commit the transaction immediately. If ``False`` the transaction will
                be committed when the connection goes out of scope and is closed (or you can
                commit manually with ``connection.commit()``).

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        with self.cursor(connection) as cursor:

            logger.debug(f'SQL Query: {sql}')
            cursor.execute(sql, parameters)

            if commit:
                connection.commit()

            # If the cursor is empty, don't cause an error
            if not cursor.description:
                logger.debug('Query returned 0 rows')
                return None

            else:

                # Fetch the data in batches, and "pickle" the rows to a temp file.
                # (We pickle rather than writing to, say, a CSV, so that we maintain
                # all the type information for each field.)

                temp_file = files.create_temp_file()

                with open(temp_file, 'wb') as f:
                    # Grab the header
                    header = [i[0] for i in cursor.description]
                    pickle.dump(header, f)

                    while True:
                        batch = cursor.fetchmany(QUERY_BATCH_SIZE)
                        if not batch:
                            break

                        logger.debug(f'Fetched {len(batch)} rows.')
                        for row in batch:
                            pickle.dump(list(row), f)

                # Load a Table from the file
                final_tbl = Table(petl.frompickle(temp_file))

                logger.debug(f'Query returned {final_tbl.num_rows} rows.')
                return final_tbl

    def _create_table_precheck(self, connection, table_name, if_exists):
        """
        Helper to determine what to do when you need a table that may already exist.

        `Args:`
            connection: obj
                A connection object obtained from ``redshift.connection()``
            table_name: str
                The table to check
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``,
                or ``truncate`` the table.
        `Returns:`
            bool
                True if the table needs to be created, False otherwise.
        """

        if if_exists not in ['fail', 'truncate', 'append', 'drop']:
            raise ValueError("Invalid value for `if_exists` argument")

        # If the table exists, evaluate the if_exists argument for next steps.
        if self.table_exists_with_connection(table_name, connection):

            if if_exists == 'fail':
                raise ValueError('Table already exists.')

            if if_exists == 'truncate':
                truncate_sql = f"TRUNCATE TABLE {table_name};"
                logger.info(f"Truncating {table_name}.")
                self.query_with_connection(truncate_sql, connection, commit=False)

            if if_exists == 'drop':
                logger.info(f"Dropping {table_name}.")
                drop_sql = f"DROP TABLE {table_name};"
                self.query_with_connection(drop_sql, connection, commit=False)
                return True

            return False

        else:
            return True

    def table_exists(self, table_name, view=True):
        """
        Check if a table or view exists in the database.

        `Args:`
            table_name: str
                The table name and schema (e.g. ``myschema.mytable``).
            view: boolean
                Check to see if a view exists by the same name

        `Returns:`
            boolean
                ``True`` if the table exists and ``False`` if it does not.
        """
        with self.connection() as connection:
            return self.table_exists_with_connection(table_name, connection, view)

    def table_exists_with_connection(self, table_name, connection, view=True):

        # Extract the table and schema from this. If no schema is detected then
        # will default to the public schema.
        try:
            schema, table = table_name.lower().split('.', 1)
        except ValueError:
            schema, table = "public", table_name.lower()

        with self.cursor(connection) as cursor:

            # Check in pg tables for the table
            sql = f"""select count(*) from pg_tables where schemaname='{schema}' and
                     tablename='{table}';"""

            cursor.execute(sql)
            result = cursor.fetchone()[0]

            # Check in the pg_views if it is a view
            if view:
                sql = f"""select count(*) from pg_views where schemaname='{schema}' and
                         viewname='{table}';"""
                cursor.execute(sql)
                result += cursor.fetchone()[0]

        # If in either, return boolean
        if result >= 1:
            return True
        else:
            return False



import requests
import os
import petl
from parsons.etl.table import Table


class TargetSmartConnector(object):

    def __init__(self, api_key=None, uri='https://api.targetsmart.com/'):

        if api_key is None:

            try:
                api_key = os.environ['TS_API_KEY']
            except KeyError:
                raise KeyError('No TargetSmart API key found. Please store'
                               ' in environment variable or pass as an'
                               'argument.')

        self.uri = uri
        self.api_key = api_key
        self.headers = {'x-api-key': self.api_key}

    def request(self, url, args=None, raw=False):

        r = requests.get(url, headers=self.headers, params=args)

        # This allows me to deal with data that needs to be munged.
        if raw:

            return r.json()

        return Table(r.json()['output'])


class Person(object):

    def __init__(self):

        return None

    def data_enhance(self, search_id, search_id_type='voterbase', state=None):
        """
        Searches for a record based on an id or phone or email address

        `Args:`
            search_id: str
                The primary key or email address or phone number
            search_id_type: str
                One of ``voterbase``, ``exacttrack``, ``abilitec_consumer_link``, ``phone``,
                ``email``, ``smartvan``, ``votebuilder``, ``voter``, ``household``.
            state: str
                Two character state code. Required if ``search_id_type`` of ``smartvan``,
                ``votebuilder`` or ``voter``.
        `Returns`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if search_id_type in ['smartvan', 'votebuilder', 'voter'] and state is None:

            raise KeyError("Search ID type '{}' requires state kwarg".format(search_id_type))

        if search_id_type not in ('voterbase', 'exacttrack', 'abilitec_consumer_link', 'phone',
                                  'email', 'smartvan', 'votebuilder', 'voter', 'household'):

            raise ValueError('Search_id_type is not valid')

        url = self.connection.uri + 'person/data-enhance'

        args = {'search_id': search_id,
                'search_id_type': search_id_type,
                'state': state
                }

        return self.connection.request(url, args=args)

    def radius_search(self, first_name, last_name, middle_name=None, name_suffix=None,
                      latitude=None, longitude=None, address=None, address_type='reg',
                      radius_size=10, radius_unit='miles', max_results=10, gender='a',
                      age_min=None, age_max=None, composite_score_min=1, composite_score_max=100,
                      last_name_exact=True, last_name_is_prefix=False, last_name_prefix_length=10):
        """
        Search for a person based on a specified radius

        `Args`:
            first_name: str
                One or more alpha characters
            last_name: str
                One or more alpha characters
            middle_name: str
                One or more alpha characters
            name_suffix: str
                One or more alpha characters
            latitude: float
                Floating point number (e.g. 33.738987255507)
            longitude: float
                Floating point number (e.g. -116.40833849559)
            address: str
                Any geocode-able address
            address_type: str
                ``reg`` for registration (default) or ``tsmart`` for TargetSmart
            radius_unit: str
                One of ``meters``, ``feet``, ``miles`` (default), or ``kilometers``.
            max_results: int
                Default of ``10``. An integer in range [0 - 100]
            gender: str
                Default of ``a``. One of ``m``, ``f``, ``u``, ``a``.
            age_min: int
                A positive integer
            age_max: int
                A positive integer
            composite_score_min: int
                An integer in range [1 - 100]. Filter out results with composite score
                less than this value.
            composite_score_max: int
                An integer in range [1 - 100]. Filter out results with composite score
                greater than this value.
            last_name_exact: boolean
                By default, the full last name is used for finding matches if the length of the
                last name is not longer than 10 characters. As an example, “anders” is less likely
                to match to “anderson” with this enabled. Disable this option if you are using
                either ``last_name_is_prefix`` or ``last_name_prefix_length``.
            last_name_is_prefix: boolean
                By default, the full last name is used for finding matches. Enable this parameter
                if your search last name is truncated. This can be common for some client
                applications that for various reasons do not have full last names. Use this
                parameter along with ``last_name_prefix_length`` to configure the length of the last
                name prefix. This parameter is ignored if ``last_name_exact`` is enabled.
            last_name_prefix_length: int
                By default, up to the first 10 characters of the search last name are used for
                finding relative matches. This value must be between 3 and 10. This parameter is
                ignored if last_name_exact is enabled.
        `Returns`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if (latitude is None or longitude is None) and address is None:
            raise ValueError('Lat/Long or Address required')

        # Convert booleans
        for a in [last_name_exact, last_name_is_prefix]:
            a = str(a)

        url = self.connection.uri + 'person/radius-search'

        args = {'first_name': first_name,
                'last_name': last_name,
                'middle_name': middle_name,
                'name_suffix': name_suffix,
                'latitude': latitude,
                'longitude': longitude,
                'address': address,
                'radius_size': radius_size,
                'radius_unit': radius_unit,
                'max_results': max_results,
                'gender': gender,
                'age_min': age_min,
                'age_max': age_max,
                'composite_score_min': composite_score_min,
                'composite_score_max': composite_score_max,
                'last_name_exact': last_name_exact,
                'last_name_is_prefix': last_name_is_prefix,
                'last_name_prefix_length': last_name_prefix_length
                }

        r = self.connection.request(url, args=args, raw=True)
        return Table([itm for itm in r['output']]).unpack_dict('data_fields', prepend=False)

    def phone(self, table):
        """
        Match based on a list of 500 phones numbers. Table
        can contain up to 500 phone numbers to match

        `Args:`
            table: parsons table
                See :ref:`parsons-table`. One row per phone number,
                up to 500 phone numbers.
        `Returns:`
            See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'person/phone-search'

        args = {'phones': list(petl.values(table.table, 0))}

        return Table(self.connection.request(url, args=args, raw=True)['result'])


class Service(object):

    def __init__(self):

        return None

    def district(self, search_type='zip', address=None, zip5=None, zip4=None, state=None,
                 latitude=None, longitude=None):
        """
        Return district information based on a geographic point. The method allows you to
        search based on the following:

        .. list-table::
            :widths: 30 30 30
            :header-rows: 1

            * - Search Type
              - Search Type Name
              - Required kwarg(s)
            * - Zip Code
              - ``zip``
              - ``zip5``, ``zip4``
            * - Address
              - ``address``
              - ``address``
            * - Point
              - point
              - ``latitude``, ``longitude``

        `Args`:
            search_type: str
                The type of district search to perform. One of ``zip``, ``address``
                or ``point``.
            address: str
                An uparsed full address
            zip5: str
                The USPS Zip5 code
            zip4: str
                The USPS Zip4 code
            state: str
                The two character state code
            latitude: float or str
                Valid latitude floating point
            lontitude: float or str
                Valid longitude floating point
        `Returns`:
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if search_type == 'zip' and None in [zip5, zip4]:
            raise ValueError("Search type 'zip' requires 'zip5' and 'zip4' arguments")

        elif search_type == 'point' and None in [latitude, longitude]:
            raise ValueError("Search type 'point' requires 'latitude' and 'longitude' arguments")

        elif search_type == 'address' and None in [address]:
            raise ValueError("Search type 'address' requires 'address' argument")

        elif search_type not in ['zip', 'point', 'address']:
            raise KeyError("Invalid 'search_type' provided. ")

        else:
            pass

        url = self.connection.uri + 'service/district'

        args = {'search_type': search_type,
                'address': address,
                'zip5': zip5,
                'zip4': zip4,
                'state': state,
                'latitude': latitude,
                'longitude': longitude
                }

        return Table([self.connection.request(url, args=args, raw=True)['match_data']])


class Voter(object):

    def __init__(self, connection):

        self.connection = connection

    def voter_registration_check(self, first_name=None, last_name=None,
                                 state=None, street_number=None,
                                 street_name=None, city=None, zip_code=None,
                                 age=None, dob=None, phone=None, email=None,
                                 unparsed_full_address=None,
                                 obj_type="dict"):
        """
        Searches for a registered individual, returns matches.

        A search must include the at minimum first name, last name and state.

        `Args:`
            first_name: str
                Required; One or more alpha characters. Trailing wildcard allowed
            last_name: str
                Required; One or more alpha characters. Trailing wildcard allowed
            state: str
                Required; Two character state code (e.g. ``NY``)
            street_number: str
                Optional; One or more alpha characters. Trailing wildcard allowed
            street_name: str
                Optional; One or more alpha characters. Trailing wildcard allowed
            city: str
                Optional; The person's home city
            zip_code: str
                Optional; Numeric characters. Trailing wildcard allowed
            age; int
                Optional; One or more integers. Trailing wildcard allowed
            dob; str
                Numeric characters in YYYYMMDD format. Trailing wildcard allowed
            phone; str
                Integer followed by 0 or more * or integers
            email: str
                Alphanumeric character followed by 0 or more * or legal characters
                (alphanumeric, @, -, .)
            unparsed_full_address: str
                One or more alphanumeric characters. No wildcards.
        `Returns`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'voter/voter-registration-check'

        if None in [first_name, last_name, state]:
            raise ValueError("""Function must include at least first_name,
                             last_name, and state.""")

        args = {'first_name': first_name,
                'last_name': last_name,
                'state': state,
                'street_number': street_number,
                'street_name': street_name,
                'city': city,
                'zip_code': zip_code,
                'age': age,
                'dob': dob,
                'phone': phone,
                'email': email,
                'unparsed_full_address': unparsed_full_address
                }

        return self.connection.request(url, args=args, raw=True)


class TargetSmartAPI(Voter, Person, Service):

    def __init__(self, api_key=None):

        self.connection = TargetSmartConnector(api_key=api_key)



from parsons.utilities import files
from parsons.utilities import check_env
import json
import os


def setup_google_application_credentials(app_creds):
    # Detect if the app_creds string is a path or json and if it is a
    # json string, then convert it to a temporary file. Then set the
    # environmental variable.
    if app_creds:
        try:
            json.loads(app_creds)
            creds_path = files.string_to_temp_file(app_creds, suffix='.json')
        except ValueError:
            creds_path = app_creds

        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = creds_path

    check_env.check('GOOGLE_APPLICATION_CREDENTIALS', app_creds)



import os


def check(env, field, optional=False):
    """
    Check if an environment variable has been set. If it has not been set
    and the passed field or arguments have not been passed, then raise an
    error.
    """
    if not field:
        try:
            return os.environ[env]
        except KeyError:
            if not optional:
                raise KeyError(f'No {env} found. Store as environment variable or '
                               f'pass as an argument.')
    return field



import os
import json
import logging

from parsons.etl.table import Table

import gspread
from oauth2client.service_account import ServiceAccountCredentials

logger = logging.getLogger(__name__)


class GoogleSheets:
    """
    A connector for Google Sheets, handling data import and export.

    `Args:`
        google_keyfile_dict: dict
            A dictionary of Google Drive API credentials, parsed from JSON provided
            by the Google Developer Console. Required if env variable
            ``GOOGLE_DRIVE_CREDENTIALS`` is not populated.
    """

    def __init__(self, google_keyfile_dict=None):

        scope = [
            'https://spreadsheets.google.com/feeds',
            'https://www.googleapis.com/auth/drive',
            ]

        self.google_keyfile_dict = google_keyfile_dict

        if google_keyfile_dict is None:
            try:
                keyfile_json = os.environ['GOOGLE_DRIVE_CREDENTIALS']
            except KeyError as error:
                logger.error("Google credentials missing. Must be specified as an env var or kwarg")
                raise error

            self.google_keyfile_dict = json.loads(keyfile_json)
        else:
            self.google_keyfile_dict = google_keyfile_dict

        credentials = ServiceAccountCredentials.from_json_keyfile_dict(
            self.google_keyfile_dict, scope
            )
        self.gspread_client = gspread.authorize(credentials)

    def _get_sheet(self, spreadsheet_id, sheet_index=0):
        return self.gspread_client.open_by_key(spreadsheet_id).get_worksheet(sheet_index)

    def get_sheet_index_with_title(self, spreadsheet_id, title):
        """
        Get the first sheet in a Google spreadsheet with the given title.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            title: str
                The sheet title

        `Returns:`
            str
                The sheet index
        """

        sheets = self.gspread_client.open_by_key(spreadsheet_id).worksheets()
        for index, sheet in enumerate(sheets):
            if sheet.title == title:
                return index
        raise ValueError(f"Couldn't find sheet with title {title}")

    def read_sheet(self, spreadsheet_id, sheet_index=0):
        """
        Create a ``parsons table`` from a sheet in a Google spreadsheet, given the sheet index.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            sheet_index: int (optional)
                The index of the desired worksheet (Index begins at 0).

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        sheet = self._get_sheet(spreadsheet_id, sheet_index)
        records = sheet.get_all_values()

        return Table(records)

    def read_sheet_with_title(self, spreadsheet_id, title):
        """
        Create a ```parsons table``` from a sheet in Google spreadsheet, given the sheet title.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            title: str
                The sheet title

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        index = self.get_sheet_index_with_title(spreadsheet_id, title)
        return self.read_sheet(spreadsheet_id, index)

    def share_spreadsheet(self, spreadsheet_id, sharee, share_type='user', role='reader',
                          notify=True, notify_message=None, with_link=False):
        """
        Share a spreadsheet with a user, group of users, domain and/or the public.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            sharee: str
                User or group e-mail address, domain name to share the spreadsheet
                with. To share publicly, set sharee value to ``None``.
            share_type: str
                The sharee type. Allowed values are: ``user``, ``group``, ``domain``,
                ``anyone``.
            role: str
                The primary role for this user. Allowed values are: ``owner``,
                ``writer``, ``reader``.
            notify: boolean
                Whether to send an email to the target user/domain.
            email_message: str
                The email to be sent if notify kwarg set to True.
            with_link: boolean
                Whether a link is required for this permission.
        """

        spreadsheet = self.gspread_client.open_by_key(spreadsheet_id)
        spreadsheet.share(sharee, share_type, role, notify=notify,
                          email_message=notify_message, with_link=with_link)

    def get_spreadsheet_permissions(self, spreadsheet_id):
        """
        List the permissioned users and groups for a spreadsheet.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        spreadsheet = self.gspread_client.open_by_key(spreadsheet_id)
        return Table(spreadsheet.list_permissions())

    def create_spreadsheet(self, title, editor_email=None):
        """
        Create a Google spreadsheet from a Parsons table. Optionally shares the new doc with
        the given email address.

        `Args:`
            title: str
                The human-readable title of the new spreadsheet
            editor_email: str (optional)
                Email address which should be given permissions on this spreadsheet

        `Returns:`
            str
                The spreadsheet ID
        """

        spreadsheet = self.gspread_client.create(title)

        if editor_email:
            self.gspread_client.insert_permission(
                spreadsheet.id,
                editor_email,
                perm_type='user',
                role='writer',
            )

        return spreadsheet.id

    def delete_spreadsheet(self, spreadsheet_id):
        """
        Deletes a Google spreadsheet.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
        """
        self.gspread_client.del_spreadsheet(spreadsheet_id)

    def add_sheet(self, spreadsheet_id, title=None, rows=100, cols=25):
        """
        Adds a sheet to a Google spreadsheet.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            rows: int
                Number of rows
            cols
                Number of cols

        `Returns:`
            str
                The sheet index
        """
        spreadsheet = self.gspread_client.open_by_key(spreadsheet_id)
        spreadsheet.add_worksheet(title, rows, cols)
        sheet_count = len(spreadsheet.worksheets())
        return (sheet_count-1)

    def append_to_sheet(self, spreadsheet_id, table, sheet_index=0, user_entered_value=False):
        """
        Append data from a Parsons table to a Google sheet. Note that the table's columns are
        ignored, as we'll be keeping whatever header row already exists in the Google sheet.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            table: obj
                Parsons table
            sheet_index: int (optional)
                The index of the desired worksheet
            user_entered_value: bool (optional)
                If True, will submit cell values as entered (required for entering formulas).
                Otherwise, values will be entered as strings or numbers only.
        """

        sheet = self._get_sheet(spreadsheet_id, sheet_index)

        # Grab the existing data, so we can figure out where to start adding new data as a batch.
        # TODO Figure out a way to do a batch append without having to read the whole sheet first.
        # Maybe use gspread's low-level batch_update().
        existing_table = self.read_sheet(spreadsheet_id, sheet_index)

        # If the existing sheet is blank, then just overwrite the table.
        if existing_table.num_rows == 0:
            return self.overwrite_sheet(spreadsheet_id, table, sheet_index, user_entered_value)

        cells = []
        for row_num, row in enumerate(table.data):
            for col_num, cell in enumerate(row):
                # Add 2 to allow for the header row, and for google sheets indexing starting at 1
                sheet_row_num = existing_table.num_rows + row_num + 2
                cells.append(gspread.Cell(sheet_row_num, col_num + 1, row[col_num]))

        value_input_option = 'RAW'
        if user_entered_value:
            value_input_option = 'USER_ENTERED'

        # Update the data in one batch
        sheet.update_cells(cells, value_input_option=value_input_option)

    def overwrite_sheet(self, spreadsheet_id, table, sheet_index=0, user_entered_value=False):
        """
        Replace the data in a Google sheet with a Parsons table, using the table's columns as the
        first row.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            table: obj
                Parsons table
            sheet_index: int (optional)
                The index of the desired worksheet
            user_entered_value: bool (optional)
                If True, will submit cell values as entered (required for entering formulas).
                Otherwise, values will be entered as strings or numbers only.
        """

        sheet = self._get_sheet(spreadsheet_id, sheet_index)
        sheet.clear()

        value_input_option = 'RAW'
        if user_entered_value:
            value_input_option = 'USER_ENTERED'

        # Add header row
        sheet.append_row(table.columns, value_input_option=value_input_option)

        cells = []
        for row_num, row in enumerate(table.data):
            for col_num, cell in enumerate(row):
                # We start at row #2 to keep room for the header row we added above
                cells.append(gspread.Cell(row_num + 2, col_num + 1, row[col_num]))

        # Update the data in one batch
        sheet.update_cells(cells, value_input_option=value_input_option)



from twilio.rest import Client
from parsons.utilities import check_env, json_format
from parsons.etl import Table
import logging


logger = logging.getLogger(__name__)


class Twilio:
    """
    `Args:`
        account_sid: str
            The Twilio account sid. Not required if ``TWILIO_ACCOUNT_SID`` env variable is
            passed.
        auth_token: str
            The Twilio auth token. Not required if ``TWILIO_AUTH_TOKEN`` env variable is
            passed.
    `Returns`:
        Twilio class
    """

    def __init__(self, account_sid=None, auth_token=None):

        self.account_sid = check_env.check('TWILIO_ACCOUNT_SID', account_sid)
        self.auth_token = check_env.check('TWILIO_AUTH_TOKEN', auth_token)
        self.client = Client(self.account_sid, self.auth_token)

    def table_convert(self, obj):
        # Internal method to create a Parsons table from a Twilio object.

        tbl = Table([x.__dict__['_properties'] for x in obj])

        if 'subresource_uris' in tbl.columns and 'uri' in tbl.columns:
            tbl.remove_column('subresource_uris', 'uri')

        return tbl

    def get_account(self, account_sid):
        """
        Get Twilio account

        `Args:`
            account_sid: str
                The Twilio account sid
        `Returns:`
            dict
        """

        r = self.client.api.accounts(account_sid)
        logger.info(f'Retrieved {account_sid} account.')
        return r.__dict__

    def get_accounts(self, name=None, status=None):
        """
        Get Twilio accounts including subaccounts.

        `Args:`
            name: str
                Filter to name of the account
            status: str
                Filter to an account status of ``active``, ``closed`` or ``suspended``.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.client.api.accounts.list(friendly_name=name, status=status)
        tbl = self.table_convert(r)

        logger.info(f'Retrieved {tbl.num_rows} accounts.')
        return tbl

    def get_account_usage(self, category=None, start_date=None, end_date=None, time_period=None,
                          group_by=None, exclude_null=False):
        """
        Get Twilio account usage.

        `Args:`
            category: str
                Filter to a specific type of usage category. The list of possibilities can be found
                `here <https://www.twilio.com/docs/usage/api/usage-record?code-sample=code-last-months-usage-for-all-usage-categories-4&code-language=Python&code-sdk-version=5.x#usage-all-categories>`_.
            start_date: str
                Filter to usage from a specific start date (ex. ``2019-01-01``).
            end_date: str
                Filter to usage from a specific end date (ex. ``2019-01-01``).
            time_period: str
                A convenience method to filter usage. Can be one of ``today``, ``yesterday``,
                ``this_month``, ``last_month``.
            group_by: str
                The time interval to group usage by. Can be one of ``daily``, ``monthly`` or
                ``yearly``.
            exclude_null: boolean
                Exclude rows that have no usage.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """ # noqa: E501,E261

        # Add populated arguments
        args = {'category': category,
                'start_date': start_date,
                'end_date': end_date}
        args = json_format.remove_empty_keys(args)

        # Parse out the time_periods
        if time_period == 'today':
            r = self.client.usage.records.today.list(**args)
        elif time_period == 'yesterday':
            r = self.client.usage.records.yesterday.list(**args)
        elif time_period == 'this_month':
            r = self.client.usage.records.this_month.list(**args)
        elif time_period == 'last_month':
            r = self.client.usage.records.last_month.list(**args)

        # Parse out the group by
        elif group_by == 'daily':
            r = self.client.usage.records.daily.list(**args)
        elif group_by == 'monthly':
            r = self.client.usage.records.monthly.list(**args)
        elif group_by == 'yearly':
            r = self.client.usage.records.yearly.list(**args)
        else:
            r = self.client.usage.records.list(**args)

        tbl = self.table_convert(r)

        if exclude_null:
            tbl.remove_null_rows('count', null_value='0')

        return tbl

    def get_messages(self, to=None, from_=None, date_sent=None, date_sent_before=None,
                     date_sent_after=None):
        """
        Get Twilio messages.

        `Args:`
            to: str
                Filter to messages only sent to the specified phone number.
            from_: str
                Filter to messages only sent from the specified phone number.
            date_sent: str
                Filter to messages only sent on the specified date (ex. ``2019-01-01``).
            date_sent_before: str
                Filter to messages only sent before the specified date (ex. ``2019-01-01``).
            date_sent_after: str
                Filter to messages only sent after the specified date (ex. ``2019-01-01``).
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.client.messages.list(to=to, from_=from_, date_sent=date_sent,
                                      date_sent_before=date_sent_before,
                                      date_sent_after=date_sent_after)

        tbl = self.table_convert(r)
        logger.info(f'Retrieved {tbl.num_rows} messages.')
        return tbl



from requests import request as _request
from requests.exceptions import HTTPError
import logging
from simplejson.errors import JSONDecodeError

logger = logging.getLogger(__name__)


class APIConnector(object):
    """
    The API Connector is a low level class for API requests that other connectors
    can utilize. It is understood that there are many standards for REST APIs and it will be
    difficult to create a universal connector. The goal of this class is create series
    of utilities that can be mixed and matched to, hopefully, meet the needs of the specific
    API.

    `Args:`
        uri: str
            The base uri for the api. Must include a trailing '/' (e.g. ``http://myapi.com/v1/``)
        headers: dict
            The request headers
        auth: dict
            The request authorization parameters
        pagination_key: str
            The name of the key in the response json where the pagination url is
            located. Required for pagination.
        data_key: str
            The name of the key in the response json where the data is contained. Required
            if the data is nested in the response json
    `Returns`:
        APIConnector class
    """

    def __init__(self, uri, headers=None, auth=None, pagination_key=None, data_key=None):

        self.uri = uri
        self.headers = headers
        self.auth = auth
        self.pagination_key = pagination_key
        self.data_key = data_key

    def request(self, url, req_type, json=None, data=None, params=None):
        """
        Base request using requests libary.

        `Args:`
            url: str
                The url request string
            req_type: str
                The request type. One of GET, POST, PATCH, DELETE, OPTIONS
            json: dict
                The payload of the request object. By using json, it will automatically
                serialize the dictionary
            data: str or byte or dict
                The payload of the request object. Use instead of json in some instances.
            params: dict
                The parameters to append to the url (e.g. http://myapi.com/things?id=1)
            raise_on_error:
                If the request yields an error status code (anything above 400), raise an
                error. In most cases, this should be True, however in some cases, if you
                are looping through data, you might want to ignore individual failures.

        `Returns:`
            requests response
        """

        return _request(req_type, url, headers=self.headers, auth=self.auth, json=json, data=data,
                        params=params)

    def get_request(self, url, params=None):
        """
        Make a GET request.

        Args:
            url: str
                A complete and valid url for the api request
            params: dict
                The request parameters
        Returns:
                A requests response object
        """

        r = self.request(url, 'GET', params=params)
        self.validate_response(r)
        logger.debug(r.json())

        return r.json()

    def post_request(self, url, params=None, data=None, json=None, success_codes=[200, 204, 201]):
        """
        Make a POST request.

        `Args:`
            url: str
                A complete and valid url for the api request
            params: dict
                The request parameters
            data: str or file
                A data object to post
            json: dict
                A JSON object to post
            success_code: int
                The expected success code to be returned
        `Returns:`
            A requests response object
        """

        r = self.request(url, 'POST', params=params, data=data, json=json)

        # Validate the response and lift up an errors.
        self.validate_response(r)

        # Check for a valid success code for the POST. Some APIs return messages with the
        # success code and some do not. Be able to account for both of these types.
        if r.status_code in success_codes:
            if self.json_check(r):
                return r.json()
            else:
                return r.status_code

    def delete_request(self, url, params=None, success_codes=[204, 201]):
        """
        Make a DELETE request.

        Args:
            url: str
                A complete and valid url for the api request
            params: dict
                The request parameters
            success_codes: int
                The expected success codes to be returned
        Returns:
                A requests response object or status code
        """

        r = self.request(url, 'DELETE', params=params)

        self.validate_response(r)

        # Check for a valid success code for the POST. Some APIs return messages with the
        # success code and some do not. Be able to account for both of these types.
        if r.status_code in success_codes:
            if self.json_check(r):
                return r.json()
            else:
                return r.status_code

    def put_request(self, url, data=None, json=None, params=None, success_codes=[204, 201]):
        """
        Make a PUT request.

        Args:
            url: str
                A complete and valid url for the api request
            params: dict
                The request parameters
            data: str or file
                A data object to post
            json: dict
                A JSON object to post
        Returns:
                A requests response object
        """

        r = self.request(url, 'PUT', params=params, data=data, json=json)

        self.validate_response(r)

        if r.status_code in success_codes:
            if self.json_check(r):
                return r.json()
            else:
                return r.status_code

    def patch_request(self, url, params=None, data=None, json=None, success_codes=[204, 201]):
        """
        Make a PATCH request.

        `Args:`
            url: str
                A complete and valid url for the api request
            params: dict
                The request parameters
            data: str or file
                A data object to post
            json: dict
                A JSON object to post
            success_codes: int
                The expected success codes to be returned
        `Returns:`
            A requests response object
        """

        r = self.request(url, 'PATCH', params=params, data=data, json=json)

        self.validate_response(r)

        # Check for a valid success code for the POST. Some APIs return messages with the
        # success code and some do not. Be able to account for both of these types.
        if r.status_code in success_codes:
            if self.json_check(r):
                return r.json()
            else:
                return r.status_code

    def validate_response(self, resp):
        """
        Validate that the response is not an error code. If it is, then raise an error
        and display the error message.

        `Args:`
            resp: object
                A response object
        """

        if resp.status_code >= 400:

            if resp.reason:
                message = f'HTTP error occurred ({resp.status_code}): {resp.reason}'
            else:
                message = f'HTTP error occurred ({resp.status_code})'

            # Some errors return JSONs with useful info about the error. Return it if exists.
            if self.json_check(resp):
                raise HTTPError(f'{message}, json: {resp.json()}')
            else:
                raise HTTPError(message)

    def data_parse(self, resp):
        """
        Determines if the response json has nested data. If it is nested, it just returns the
        data. This is useful in dealing with requests that might return multiple records, while
        others might return only a single record.

        `Args:`
            resp:
                A response dictionary
        `Returns:`
            dict
                A dictionary of data.
        """

        # To Do: Some response jsons are enclosed in a list. Need to deal with unpacking and/or
        # not assuming that it is going to be a dict.

        # In some instances responses are just lists.
        if isinstance(resp, list):
            return resp

        if self.data_key and self.data_key in resp.keys():
            return resp[self.data_key]
        else:
            return resp

    # There are many different ways in which APIs indicate whether there is a next page
    # of data following the initial request. The goal is build out a series of utilities
    # that mean most of the most common use cases.

    def next_page_check_url(self, resp):
        """
        Check to determine if there is a next page. This requires that the response json
        contains a pagination key that is empty if there is not a next page.

        `Args:`
            resp:
                A response dictionary
        `Returns:
            boolean
        """

        if self.pagination_key and self.pagination_key in resp.keys():
            if resp[self.pagination_key]:
                return True
        else:
            return False

    def json_check(self, resp):
        """
        Check to see if a response has a json included in it.
        """

        try:
            resp.json()
            return True
        except JSONDecodeError:
            return False



import os
import unittest
import requests_mock
from test.utils import assert_matching_tables

from parsons.etl.table import Table
from parsons.zoom.zoom import Zoom


API_KEY = 'fake_api_key'
API_SECRET = 'fake_api_secret'
ZOOM_URI = 'https://api.zoom.us/v2/'


class TestZoom(unittest.TestCase):

    def setUp(self):

        self.zoom = Zoom(API_KEY, API_SECRET)

    @requests_mock.Mocker()
    def test_get_users(self, m):

      user_json = {
                   'page_count': 1,
                   'page_number': 1,
                   'page_size': 30,
                   'total_records': 1,
                   'users': [{
                      'id': 'C5A2nRWwTMm_hXyJb1JXMh',
                      'first_name': 'Bob',
                      'last_name': 'McBob',
                      'email': 'bob@bob.com',
                      'type': 2,
                      'pmi': 8374523641,
                      'timezone': 'America/New_York',
                      'verified': 1,
                      'dept': '',
                      'created_at': '2017-10-06T15:22:34Z',
                      'last_login_time': '2020-05-06T16:50:45Z',
                      'last_client_version': '',
                      'language': '',
                      'phone_number': '',
                      'status': 'active'}]}

      tbl = Table([{'id': 'C5A2nRWwTMm_hXyJb1JXMh',
                    'first_name': 'Bob',
                    'last_name': 'McBob',
                    'email': 'bob@bob.com',
                    'type': 2,
                    'pmi': 8374523641,
                    'timezone': 'America/New_York',
                    'verified': 1,
                    'dept': '',
                    'created_at': '2017-10-06T15:22:34Z',
                    'last_login_time': '2020-05-06T16:50:45Z',
                    'last_client_version': '',
                    'language': '',
                    'phone_number': '',
                    'status': 'active'}])

      m.get(ZOOM_URI + 'users', json=user_json)
      assert_matching_tables(self.zoom.get_users(), tbl)

    @requests_mock.Mocker()
    def test_get_meeting_participants(self, m):

      participants = {
                     'page_count': 1,
                     'page_size': 30,
                     'total_records': 4,
                     'next_page_token': '',
                     'participants': [{
                          'id': '',
                          'user_id': '16778240',
                          'name': 'Barack Obama',
                          'user_email': '',
                          'join_time': '2020-04-24T21:00:26Z',
                          'leave_time': '2020-04-24T22:24:38Z',
                          'duration': 5052,
                      'attentiveness_score': ''
                  }, {
                      'id': '',
                      'user_id': '16779264',
                      'name': '',
                      'user_email': '',
                      'join_time': '2020-04-24T21:00:45Z',
                      'leave_time': '2020-04-24T22:24:38Z',
                      'duration': 5033,
                      'attentiveness_score': ''
                  }]}

      tbl = Table([{
                   'id': '',
                   'user_id': '16778240',
                   'name': 'Barack Obama',
                   'user_email': '',
                   'join_time': '2020-04-24T21:00:26Z',
                   'leave_time': '2020-04-24T22:24:38Z',
                   'duration': 5052,
                   'attentiveness_score': ''
                    }, {
                   'id': '',
                   'user_id': '16779264',
                   'name': '',
                   'user_email': '',
                   'join_time': '2020-04-24T21:00:45Z',
                   'leave_time': '2020-04-24T22:24:38Z',
                   'duration': 5033,
                   'attentiveness_score': ''}])

      m.get(ZOOM_URI + '/report/meetings/123/participants', json=participants)
      assert_matching_tables(self.zoom.get_past_meeting_participants(123), tbl)


import dateutil.parser as dp


def iso_to_unix(iso_date):
    # Convert an ISO date to a Unix time

    if iso_date:
        parsed_date = dp.parse(iso_date)
        return int(parsed_date.strftime('%s'))
    else:
        return None



import unittest
import os
import requests_mock
from parsons.ngpvan.van import VAN
from test.utils import validate_list, assert_matching_tables
from requests.exceptions import HTTPError


custom_field = [{
    "customFieldId": 157,
    "customFieldParentId": None,
    "customFieldName": "Education level",
    "customFieldGroupId": 52,
    "customFieldGroupName": "Education",
    "customFieldGroupType": "Contacts",
    "customFieldTypeId": "S",
    "isEditable": True,
    "isExportable": False,
    "maxTextboxCharacters": None,
    "availableValues": [
      {"id": 1, "name": "High School diploma", "parentValueId": None},
      {"id": 2, "name": "College degree", "parentValueId": None},
      {"id": 3, "name": "Postgraduate degree", "parentValueId": None},
      {"id": 4, "name": "Doctorate", "parentValueId": None}]}]

custom_field_values = [
  {'customFieldId': 157, 'id': 1, 'name': 'High School diploma', 'parentValueId': None},
  {'customFieldId': 157, 'id': 2, 'name': 'College degree', 'parentValueId': None},
  {'customFieldId': 157, 'id': 3, 'name': 'Postgraduate degree', 'parentValueId': None},
  {'customFieldId': 157, 'id': 4, 'name': 'Doctorate', 'parentValueId': None}
  ]

os.environ['VAN_API_KEY'] = 'SOME_KEY'


class TestCustomFields(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters")

    @requests_mock.Mocker()
    def test_get_custom_fields(self, m):

        m.get(self.van.connection.uri + 'customFields', json=custom_field)
        assert_matching_tables(custom_field, self.van.get_custom_fields())

    @requests_mock.Mocker()
    def test_get_custom_field_values(self, m):

        m.get(self.van.connection.uri + 'customFields', json=custom_field)
        assert_matching_tables(custom_field_values, self.van.get_custom_fields_values())

    @requests_mock.Mocker()
    def test_get_custom_field(self, m):

        m.get(self.van.connection.uri + 'customFields/157', json=custom_field)
        assert_matching_tables(custom_field, self.van.get_custom_field(157))



from parsons.aws.s3 import S3
from parsons.aws.lambda_distribute import distribute_task
from parsons.aws.aws_async import event_command

__all__ = [
    'S3',
    'distribute_task',
    'event_command'
]



import ast
import petl
import logging

logger = logging.getLogger(__name__)

# These are reserved words by Postgres/Redshift and cannot be used as column names.
RESERVED_WORDS = ['AES128', 'AES256', 'ALL', 'ALLOWOVERWRITE', 'ANALYSE', 'ANALYZE', 'AND', 'ANY',
                  'ARRAY', 'AS', 'ASC', 'AUTHORIZATION', 'BACKUP', 'BETWEEN', 'BINARY',
                  'BLANKSASNULL', 'BOTH', 'BYTEDICT', 'BZIP2', 'CASE', 'CAST', 'CHECK', 'COLLATE',
                  'COLUMN', 'CONSTRAINT', 'CREATE', 'CREDENTIALS', 'CROSS', 'CURRENT_DATE',
                  'CURRENT_TIME', 'CURRENT_TIMESTAMP', 'CURRENT_USER', 'CURRENT_USER_ID',
                  'DEFAULT', 'DEFERRABLE', 'DEFLATE', 'DEFRAG', 'DELTA', 'DELTA32K', 'DESC',
                  'DISABLE', 'DISTINCT', 'DO', 'ELSE', 'EMPTYASNULL', 'ENABLE', 'ENCODE', 'ENCRYPT',
                  'ENCRYPTION', 'END', 'EXCEPT', 'EXPLICIT', 'FALSE', 'FOR', 'FOREIGN', 'FREEZE',
                  'FROM', 'FULL', 'GLOBALDICT256', 'GLOBALDICT64K', 'GRANT', 'GROUP', 'GZIP',
                  'HAVING', 'IDENTITY', 'IGNORE', 'ILIKE', 'IN', 'INITIALLY', 'INNER', 'INTERSECT',
                  'INTO', 'IS', 'ISNULL', 'JOIN', 'LEADING', 'LEFT', 'LIKE', 'LIMIT', 'LOCALTIME',
                  'LOCALTIMESTAMP', 'LUN', 'LUNS', 'LZO', 'LZOP', 'MINUS', 'MOSTLY13', 'MOSTLY32',
                  'MOSTLY8', 'NATURAL', 'NEW', 'NOT', 'NOTNULL', 'NULL', 'NULLS', 'OFF', 'OFFLINE',
                  'OFFSET', 'OLD', 'ON', 'ONLY', 'OPEN', 'OR', 'ORDER', 'OUTER', 'OVERLAPS',
                  'PARALLEL', 'PARTITION', 'PERCENT', 'PERMISSIONS', 'PLACING', 'PRIMARY', 'RAW',
                  'READRATIO', 'RECOVER', 'REFERENCES', 'RESPECT', 'REJECTLOG', 'RESORT', 'RESTORE',
                  'RIGHT', 'SELECT', 'SESSION_USER', 'SIMILAR', 'SOME', 'SYSDATE', 'SYSTEM',
                  'TABLE', 'TAG', 'TDES', 'TEXT255', 'TEXT32K', 'THEN', 'TIMESTAMP', 'TO', 'TOP',
                  'TRAILING', 'TRUE', 'TRUNCATECOLUMNS', 'UNION', 'UNIQUE', 'USER', 'USING',
                  'VERBOSE', 'WALLET', 'WHEN', 'WHERE', 'WITH', 'WITHOUT']


class PostgresCreateStatement(object):

    def __init__(self):

        pass

    def create_statement(self, tbl, table_name, padding=None, distkey=None, sortkey=None,
                         varchar_max=None, varchar_truncate=True, columntypes=None):
        # Generate a table create statement. Distkeys and sortkeys are only used by
        # Redshift and should not be passed when generating a create statement for
        # Postgres.

        if tbl.num_rows == 0:
            raise ValueError('Table is empty. Must have 1 or more rows.')

        # Validate and rename column names if needed
        tbl.table = petl.setheader(tbl.table, self.column_name_validate(tbl.columns))

        mapping = self.generate_data_types(tbl)

        if padding:
            mapping['longest'] = self.vc_padding(mapping, padding)

        if varchar_max:
            mapping['longest'] = self.vc_max(mapping, varchar_max)

        if varchar_truncate:
            mapping['longest'] = self.vc_trunc(mapping)

        mapping['longest'] = self.vc_validate(mapping)

        # Add any provided column type overrides
        if columntypes:
            for i in range(len(mapping['headers'])):
                col = mapping['headers'][i]
                if columntypes.get(col):
                    mapping['type_list'][i] = columntypes[col]

        # Enclose in quotes
        mapping['headers'] = [f'"{h}"'for h in mapping['headers']]

        return self.create_sql(table_name, mapping, distkey=distkey, sortkey=sortkey)

    def data_type(self, val, current_type):
        # Determine the database data type of a given value

        try:
            # Convert to string to reevaluate data type
            t = ast.literal_eval(str(val))
        except ValueError:
            return 'varchar'
        except SyntaxError:
            return 'varchar'

        if type(t) in [int, float]:
            if (type(t) in [int] and current_type not in ['float', 'varchar']):

                # Make sure that it is a valid integer
                if not self.is_valid_integer(val):
                    return 'varchar'

                # Use smallest possible int type
                if (-32768 < t < 32767) and current_type not in ['int', 'bigint']:
                    return 'smallint'
                elif (-2147483648 < t < 2147483647) and current_type not in ['bigint']:
                    return 'int'
                else:
                    return 'bigint'
            if type(t) is float and current_type not in ['varchar']:
                return 'decimal'
        else:
            return 'varchar'

    def is_valid_integer(self, val):

        # Valid ints in python can contain an underscore, but Postgres can't. This
        # checks to see if there is an underscore in the value and turns it into
        # a varchar if so.
        try:
            if '_' in val:
                return False

        except TypeError:
            return True

        # If it has a leading zero, we should treat it as a varchar, since it is
        # probably there for a good reason (e.g. zipcode)
        if val.isdigit():
            if val[0] == '0':
                return False

        return True

    def generate_data_types(self, table):
        # Generate column data types

        longest, type_list = [], []

        cont = petl.records(table.table)

        # Populate empty values for the columns
        for col in table.columns:
            longest.append(0)
            type_list.append('')

        for row in cont:
            for i in range(len(row)):
                # NA is the csv null value
                if type_list[i] == 'varchar' or row[i] == 'NA':
                    pass
                else:
                    var_type = self.data_type(row[i], type_list[i])
                    type_list[i] = var_type
                # Calculate width
                width = len(str(row[i]).encode('utf-8'))
                if width > longest[i]:
                    longest[i] = width

        return {'longest': longest,
                'headers': table.columns,
                'type_list': type_list}

    def vc_padding(self, mapping, padding):
        # Pad the width of a varchar column

        return [int(c + (c * padding)) for c in mapping['longest']]

    def vc_max(self, mapping, columns):
        # Set the varchar width of a column to the maximum

        for c in columns:

            try:
                idx = mapping['headers'].index(c)
                mapping['longest'][idx] = 65535

            except KeyError as error:
                logger.error('Could not find column name provided.')
                raise error

        return mapping['longest']

    def vc_trunc(self, mapping):

        return [65535 if c > 65535 else c for c in mapping['longest']]

    def vc_validate(self, mapping):

        return [1 if c == 0 else c for c in mapping['longest']]

    def create_sql(self, table_name, mapping, distkey=None, sortkey=None):
        # Generate the sql to create the table

        statement = f'create table {table_name} ('

        for i in range(len(mapping['headers'])):
            if mapping['type_list'][i] == 'varchar':
                statement = (statement + '\n  {} varchar({}),').format(str(mapping['headers'][i])
                                                                       .lower(),
                                                                       str(mapping['longest'][i]))
            else:
                statement = (statement + '\n  ' + '{} {}' + ',').format(str(mapping['headers'][i])
                                                                        .lower(),
                                                                        mapping['type_list'][i])

        statement = statement[:-1] + ') '

        if distkey:
            statement += f'\ndistkey({distkey}) '

        if sortkey:
            statement += f'\nsortkey({sortkey})'

        statement += ';'

        return statement

    def column_name_validate(self, columns):
        # Validate the column names and rename if not valid

        clean_columns = []

        for idx, c in enumerate(columns):

            # Remove spaces. Technically allowed with double quotes
            # but I think that it is bad practice.
            c = c.replace(' ', '')

            # if column is an empty string, replace with 'col_INDEX'
            if c == '':
                logger.info(f'Column is an empty string. Renaming column.')
                c = f'col_{idx}'

            # If column is a reserved word, replace with 'col_INDEX'. Technically
            # you can allow these with quotes, but I think that it is bad practice
            if c.upper()in RESERVED_WORDS:
                logger.info(f'{c} is a Redshift/Postgres reserved word. Renaming column.')
                c = f'col_{idx}'

            # If column name begins with an integer, preprent with 'x_'
            if c[0].isdigit():
                logger.info(f'{c} begins with digit. Renaming column.')
                c = f'x_{c}'

            # If column name length is greater than 120 characters, truncate.
            # Technically, you can have up to 127 bytes, which might allow
            # for a few more characters, but playing it safe.
            if len(c) > 120:
                logger.info(f'Column {c[:10]}... too long. Truncating column name.')
                c = c[:120]

            clean_columns.append(c)

        return clean_columns



# Adapted from Gmail API tutorial https://developers.google.com/gmail/api
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.image import MIMEImage
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.application import MIMEApplication
from email.encoders import encode_base64
from email.utils import parseaddr
from validate_email import validate_email
import io
import logging
import mimetypes
import os

# BUG: can't send files equal to or larger than 6MB
# There is a possible fix
# `BUG-1` are the changes that would need to be made

# BUG-1
# import httplib2shim
# import sys

logger = logging.getLogger(__name__)


class SendMail(object):
    """SendMail base class for sending emails.
    """

    log = logger

    def _send_message(self, message):
        raise NotImplementedError("send_message is how to send the prepared message")

    def _create_message_simple(self, sender, to, subject, message_text):
        """Create a text-only message for an email.

        `Args:`
            sender: str
                Email address of the sender.
            to: str
                Email address(es) of the recipient(s).
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
        `Returns:`
            An object passable to send_message to send
        """
        self.log.info("Creating a simple message...")

        message = MIMEText(message_text)
        message['to'] = to
        message['from'] = sender
        message['subject'] = subject

        return message

    def _create_message_html(self, sender, to, subject, message_text,
                             message_html):
        """Create an html message for an email.

        `Args:`
            sender: str
                Email address of the sender.
            to: str
                Email address(es) of the recipient(s).
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
            message_html: str
                The html formatted text of the email message.
        `Returns:`
            An object passable to send_message to send
        """
        self.log.info("Creating an html message...")

        message = MIMEMultipart('alternative')
        message['subject'] = subject
        message['from'] = sender
        message['to'] = to
        if message_text:
            message.attach(MIMEText(message_text, 'plain'))
        message.attach(MIMEText(message_html, 'html'))

        return message

    def _create_message_attachments(self, sender, to, subject, message_text,
                                    files, message_html=None):
        """Create a message for an email that includes an attachment.

        `Args:`
            sender: str
                Email address of the sender.
            to: str
                Email address of the receiver.
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
            files: list
                The path(s) to the file(s) to be attached.
            message_html: str
                Optional; The html formatted text of the email message.
        `Returns:`
            An object passable to send_message to send
        """
        self.log.info("Creating a message with attachments...")

        message = MIMEMultipart('alternative')
        message['to'] = to
        message['from'] = sender
        message['subject'] = subject

        msg = MIMEText(message_text, 'plain')
        message.attach(msg)

        if message_html:
            html = MIMEText(message_html, 'html')
            message.attach(html)

        for f in files:
            filename = getattr(f, 'name', 'file')
            file_bytes = b''

            if isinstance(f, io.StringIO):
                file_bytes = f.getvalue().encode()
            elif isinstance(f, io.BytesIO):
                file_bytes = f.getvalue()
            else:
                filename = os.path.basename(f)
                fp = open(f, 'rb')
                file_bytes = fp.read()
                fp.close()

            content_type, encoding = mimetypes.guess_type(filename)
            self.log.debug(
                f"(File: {f}, Content-type: {content_type}, "
                f"Encoding: {encoding})")

            if content_type is None or encoding is not None:
                content_type = 'application/octet-stream'

            main_type, sub_type = content_type.split('/', 1)

            if main_type == 'text':
                self.log.info("Added a text file.")
                msg = MIMEText(file_bytes, _subtype=sub_type, _charset='utf-8')

            elif main_type == 'image':
                self.log.info("Added an image file.")
                msg = MIMEImage(file_bytes, _subtype=sub_type)
                msg.add_header('Content-ID', f'<{filename}>')

            elif main_type == 'audio':
                self.log.info("Added an audio file.")
                msg = MIMEAudio(file_bytes, _subtype=sub_type)

            elif main_type == 'application':
                self.log.info("Added an application file.")
                msg = MIMEApplication(file_bytes, _subtype=sub_type)

            else:
                self.log.info("Added an unknown-type file.")
                msg = MIMEBase(main_type, sub_type)
                msg.set_payload(file_bytes)
                encode_base64(msg)

            msg.add_header(
                'Content-Disposition', 'attachment', filename=filename)
            message.attach(msg)

        return message

    def _validate_email_string(self, str):
        self.log.debug(f"Validating email {str}...")
        realname, email_addr = parseaddr(str)

        if not email_addr:
            raise ValueError("Invalid email address.")

        if not validate_email(email_addr):
            raise ValueError("Invalid email address.")

        return True

    def send_email(self, sender, to, subject, message_text, message_html=None,
                   files=None):
        """Send an email message.

        `Args:`
            sender: str
                Email address of the sender.
            to: str or list
                Email address(es) of the receiver(s). Must be in correct email
                string syntax. For example, `name@email.com` or
                `"Name" <email@email.com>`.
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
            message_html: str
                The html formatted text of the email message. If ommitted, the
                email is sent a text-only body.
            files: str or list
                The path to the file(s) to be attached.

        `Returns:`
            None
        """
        self.log.info("Preparing to send and email...")

        self.log.info("Validating email(s)")
        if isinstance(to, list):
            if len(to) == 0:
                raise EmptyListError("Must contain at least 1 email.")

            for e in to:
                self._validate_email_string(e)

            to = ', '.join(to)

        elif isinstance(to, str):
            self._validate_email_string(to)

        if not message_html and not files:
            msg_type = 'simple'
            msg = self._create_message_simple(sender, to, subject, message_text)

        elif not files:
            msg_type = 'html'
            msg = self._create_message_html(
                sender, to, subject, message_text, message_html)
        else:
            msg_type = 'attachments'
            if isinstance(files, str):
                files = [files]

            msg = self._create_message_attachments(
                sender, to, subject, message_text, files, message_html)

        self.log.info(f"Sending a(n) {msg_type} email...")

        self._send_message(msg)

        self.log.info("Email sent succesfully.")


class EmptyListError(IndexError):
    """Throw when a list is empty that should contain at least 1 element."""
    pass



import base64
from apiclient import errors
from googleapiclient.discovery import build
from httplib2 import Http
from oauth2client import file, client, tools
from parsons.notifications.sendmail import SendMail

SCOPES = 'https://www.googleapis.com/auth/gmail.send'


class Gmail(SendMail):
    """Create a Gmail object, for sending emails.

    `Args:`
        creds_path: str
            The path to the credentials.json file.
        token_path: str
            The path to the token.json file.
        user_id: str
            Optional; Sender email address. Defaults to the special value
            "me" which is used to indicate the authenticated user.
    """

    def __init__(self, creds_path=None, token_path=None, user_id='me'):

        self.user_id = user_id

        if not creds_path:
            raise ValueError("Invalid path to credentials.json.")

        if not token_path:
            raise ValueError("Invalid path to token.json.")

        self.store = file.Storage(token_path)
        self.creds = self.store.get()

        # BUG-1
        # http = httplib2shim.Http()

        if not self.creds or self.creds.invalid:
            flow = client.flow_from_clientsecrets(creds_path, SCOPES)
            self.creds = tools.run_flow(flow, self.store)

            # BUG-1
            # self.creds = self.run_flow(flow, self.store, http=http)

        self.service = build('gmail', 'v1', http=self.creds.authorize(Http()))

        # BUG-1
        # self.service = build('gmail', 'v1', http=self.creds.authorize(http))

    def _encode_raw_message(self, message):
        return {'raw': base64.urlsafe_b64encode(message.as_bytes()).decode()}

    def _send_message(self, msg):
        """Send an email message.

        `Args:`
            message: dict
                Message to be sent as a base64url encode object.
                i.e. the objects created by the create_* instance methods
        `Returns:`
            dict
                A Users.messages object see `https://developers.google.com/gmail/api/v1/reference/users/messages#resource.` # noqa
                for more info.
        """
        self.log.info("Sending a message...")

        message = self._encode_raw_message(msg)

        self.log.debug(message)

        try:
            message = (self.service.users().messages()
                       .send(userId=self.user_id, body=message).execute())
        except errors.HttpError:
            self.log.exception(
                'An error occurred: while attempting to send a message.')
            raise
        else:
            self.log.debug(message)
            self.log.info(
                f"Message sent succesfully (Message Id: {message['id']})")

            return message



import errno
import gzip
import os
import tempfile

__all__ = [
    'create_temp_file',
    'create_temp_file_for_path',
    'is_gzip_path',
    'suffix_for_compression_type',
    'compression_type_for_path',
    'string_to_temp_file'
    ]


# Maximum number of times to try to open a new temp file before giving up.
TMP_MAX = 1000


# This global list keeps track of all temp files created during the runtime of a script.
# We can't rely exclusively on the "automatic removal" behavior of the built-in `tempfile`
# library, because of our use of petl. Specifically, if a petl table is loaded from a
# temporary file (eg. a CSV), petl may not actually read the file until much later, after the
# TemporaryFile object has already gone out of scope and the file removed. If this
# occurs, the petl load will fail since it's trying to read from a file that doesn't exist.
# So by storing all the temp files in this list, we can keep them "in scope" for the remaining
# runtime of the script.
_temp_files = []


def create_temp_file(suffix=None):
    """
    Create a temp file that will exist as long as the current script is running.

    `Args:`
        suffix: str
            A suffix/extension to add to the end of the temp file name
    `Returns:`
        str
            The path of the temp file
    """
    temp_file = TempFile(suffix=suffix)
    _temp_files.append(temp_file)
    return temp_file.name


def create_temp_file_for_path(path):
    """
    Creates a temp file that will exist as long as the current script is running, and with
    a file name mimicking that of the provided path.

    `Args:`
        path: str
            Path (or just file name) of the file you want the temp file to mimick.
    `Returns:`
        str
            The path of the temp file
    """

    # Add the appropriate compression suffix to the file, so other libraries that check the
    # file's extension will know that it is compressed.
    # TODO Make this more robust, maybe even using the entire remote file name as the suffix.
    suffix = '.gz' if is_gzip_path(path) else None
    return create_temp_file(suffix=suffix)


def close_temp_file(path):
    """
    Force closes a Parsons temp file, which will cause it to be deleted immediately.

    Useful for when you don't want to wait until the end of your script's execution for temp
    files to be closed and deleted. Eg. If you're running into system limits on open file
    descriptors.

    `Args:`
        path: str
            Path of a temp file created by ``create_temp_file``
    `Returns:`
        bool
            Whether the temp file was found and closed
    """

    for temp_file in _temp_files:
        if temp_file.name == path:
            # Call close explicitly to clean up, because we can't always assume that de-refencing
            # will necessarily result in the TempFilePath being cleaned up (depends on platform)
            temp_file.remove()
            _temp_files.remove(temp_file)
            return True

    return False


def is_gzip_path(path):
    return (path[-3:] == '.gz')


def is_zip_path(path):
    return (path[-4:] == '.zip')


def suffix_for_compression_type(compression):
    if compression == 'gzip':
        return '.gz'

    return ''


def compression_type_for_path(path):
    if is_gzip_path(path):
        return 'gzip'

    if is_zip_path(path):
        return 'zip'

    return None


def read_file(path):
    """
    Return the contents of file. Currently support `.gz` compressed files.

    `Args:`
        path: str
            The path to the file to read.
    `Returns:`
        str
            The contents of a files.
    """
    compression = compression_type_for_path(path)

    open_func = {
        'gzip': gzip.open,
        None: open,
    }
    with open_func[compression](path, 'r') as fp:
        return fp.read()


def string_to_temp_file(string, suffix=None):
    """
    Create a temporary file from a string. Currently used for packages
    that require credentials to be stored as a file.
    """

    temp_file_path = create_temp_file(suffix=suffix)

    with open(temp_file_path, 'w') as f:
        f.write(string)

    return temp_file_path


def zip_check(file_path, compression_type):
    """
    Check if the file suffix or the compression type indicates that it is
    a zip file.
    """

    if file_path:
        if file_path.split('/')[-1].split('.')[-1] == 'zip':
            return True

    if compression_type == 'zip':
        return True

    else:
        return False


def extract_file_name(file_path=None, include_suffix=True):
    """
    Extract the file name with the file path string.

    file_path: str
        The file path
    include_suffix: boolean
        If True, includes full file name with suffix. If False returns the
        file name without the suffix (e.g. "myfile.zip" vs. "myfile").
    """

    if not file_path:
        return None

    if include_suffix:
        return file_path.split('/')[-1]

    return file_path.split('/')[-1].split('.')[0]


def has_data(file_path):
    """
    Check if a file has any data in it.

    `Args:`
        file_path: str
            The file path.
    `Returns:`
        boolean
            ``True`` if data in the file and ``False`` if not.
    """

    if os.stat(file_path).st_size == 0:
        return False

    else:
        return True


def generate_tempfile(suffix):
    """
    Create a new temp file with a unique filename.

    `Args:`
        suffix: str
            The suffix to give the file path in order to advertise the file/mime type of the file.
    `Returns`
        str
            The path of the newly created temp file.
    """
    # _get_candidate_names gives us an iterator that will keep trying to generate a random filename.
    # It's not ideal to use a "protected" function from another module, but this function does some
    # heavy lifting for us.
    names = tempfile._get_candidate_names()
    temp_dir = tempfile.gettempdir()

    # Try multiple times to create a temp file, just in case (however unlikely) we have some
    # collisions with already existing files.
    for _ in range(TMP_MAX):
        name = next(names)
        if suffix:
            name = f'{name}{suffix}'
        path = os.path.join(temp_dir, name)

        try:
            # "Touch" the file to ensure that there is a file there, so that if our user tries
            # open it in read mode later, they won't get an error about the file not existing.
            # Also, use mode='x' (exclusive create) to make sure we get an error if the file already
            # exists
            with open(path, mode='x') as _:
                pass
            return path
        # PermissionError can be Windows' way of saying the file exists
        except (FileExistsError, PermissionError):
            continue    # try again with another filename if we got an error

    raise FileExistsError(errno.EEXIST,
                          "No usable temporary directory name found")


class TempFile:
    """
    Class for creating and eventually cleaning up a temporary file.

    Creating the instance of the TempFilePath will create a uniquely named temporary file. When the
    instance is garbage collected (e.g., when the Python process closes) or when the remove method
    is called explicitly, the temporary file is removed from disk.

    Unlike NamedTemporaryFile from the Python standard library, this class does NOT represent
    an open file handle to the file. It simply represents a file on disk. This class was
    written to workaround the fact that on Windows, NamedTemporaryFile opens the file with an
    exclusive read lock, which means that no one else can open the file for reading.

    Since Parsons hands out the temporary file's path and not the file handle, users must be able
    to open the file, but that is impossible as long as NamedTemporaryFile holds onto the open
    file handle with its exclusive read lock. So we wrote, TempFilePath to not hold onto the open
    file handle.

    `Args:`
        suffix: str
            The suffix to give the file path in order to advertise the file/mime type of the file.
    """

    def __init__(self, suffix=None):
        self.remove_called = False
        self.name = generate_tempfile(suffix)

    def __del__(self):
        # When we are being cleaned up, call remove to make sure the file is removed from disk.
        self.remove()

    def remove(self, unlink=os.unlink):
        """
        Remove the file from disk.

        Note: We cache a reference to the os.unlink function because during shutdown of the Python
        process, the reference to the os module may be None'd out as part of garbage collection.
        So, we want to make sure we have a reference to the function saved somewhere.

        `Args:`
            unlink: function
                Function to use for removing the file from disk.
        """
        # Only try to unlink if we have a valid file path and we haven't yet called close.
        if self.name and not self.remove_called:
            try:
                unlink(self.name)
            except FileNotFoundError:
                pass  # if the file isn't found, our work is done

        self.remove_called = True




def arg_format(arg):
    """
    Many APIs require arguments to formatted like this 'thisTypeConfig' which is not the standard
    for python so this method takes an argument 'this_type_config' and returns it as
    'thisTypeConfig'
    """

    arg_list = arg.split('_')
    arg_list = [a.capitalize() for a in arg_list]
    arg_list[0] = arg_list[0].lower()

    return ''.join(arg_list)


def remove_empty_keys(dirty_dict):
    """
    Remove empty keys from a dictionary. This method is useful when passing jsons
    in which a null field will update the value to null and you don't want that.
    """

    clean_dict = {}

    for k, v in dirty_dict.items():
        if v:
            clean_dict[k] = v

    return clean_dict


def flatten_json(json):
    """
    Flatten nested json to return a dict without nested values. Lists without nested values will
    be ignored, and lists of dicts will only return the first key value pair for each key. Useful
    for passing nested json to validation methods.
    """
    out = {}

    def flatten(x, name=''):
        if type(x) is dict:
            for k, v in x.items():
                flatten(v, k)
        elif type(x) is list:
            for a in x:
                flatten(a)
        elif name != '' and name not in out:
            out[name] = x

    flatten(json)
    return out



import logging

logger = logging.getLogger(__name__)


class DBSync:
    """
    Sync tables between databases. Works with ``Postgres``, ``Redshift``, ``MySQL``
    databases.

    `Args:`
        source_db: Database connection object
            A database object.
        destination_db: Database connection object
            A database object.
        chunk_size: int
            The number of rows per transaction copy when syncing a table. The
            default value is 100,000 rows.
    `Returns:`
        A DBSync object.
    """

    def __init__(self, source_db, destination_db, chunk_size=100000):

        self.source_db = source_db
        self.dest_db = destination_db
        self.chunk_size = chunk_size

    def table_sync_full(self, source_table, destination_table, if_exists='drop',
                        **kwargs):
        """
        Full sync of table from a source database to a destination database. This will
        wipe all data from the destination table.

        `Args:`
            source_table: str
                Full table path (e.g. ``my_schema.my_table``)
            destination_table: str
                Full table path (e.g. ``my_schema.my_table``)
            if_exists: str
                If destination table exists either ``drop`` or ``truncate``. Truncate is
                useful when there are dependent views associated with the table.
            **kwargs: args
                Optional copy arguments for destination database.
        `Returns:`
            ``None``
        """

        # Create the table objects
        source_tbl = self.source_db.table(source_table)
        destination_tbl = self.dest_db.table(destination_table)

        logger.info(f'Syncing full table data from {source_table} to {destination_table}')

        # Drop or truncate if the destination table exists
        if destination_tbl.exists:
            if if_exists == 'drop':
                destination_tbl.drop()
            elif if_exists == 'truncate':
                self._check_column_match(source_tbl, destination_tbl)
                destination_tbl.truncate()
            else:
                raise ValueError('Invalid if_exists argument. Must be drop or truncate.')

        # If the source table contains 0 rows, do not attempt to copy the table.
        if source_tbl.num_rows == 0:
            logger.info('Source table contains 0 rows')
            return None

        # Copy rows in chunks.
        copied_rows = 0
        while copied_rows < source_tbl.num_rows:

            # Get a chunk
            rows = source_tbl.get_rows(offset=copied_rows, chunk_size=self.chunk_size)

            # Copy the chunk
            self.dest_db.copy(rows, destination_table, if_exists='append', **kwargs)

            # Update counter
            copied_rows += rows.num_rows

        logger.info(f'{source_table} synced: {copied_rows} total rows copied.')

        self._row_count_verify(source_tbl, destination_tbl)

    def table_sync_incremental(self, source_table, destination_table, primary_key,
                               distinct_check=True, **kwargs):
        """
        Incremental sync of table from a source database to a destination database
        using an incremental primary key.

        `Args:`
            source_table: str
                Full table path (e.g. ``my_schema.my_table``)
            destination_table: str
                Full table path (e.g. ``my_schema.my_table``)
            if_exists: str
                If destination table exists either ``drop`` or ``truncate``. Truncate is
                useful when there are dependent views associated with the table.
            primary_key: str
                The name of the primary key. This must be the same for the source and
                destination table.
            distinct_check: bool
                Check that the source table primary key is distinct prior to running the
                sync. If it is not, an error will be raised.
            **kwargs: args
                Optional copy arguments for destination database.
        `Returns:`
            ``None``
        """

        # Create the table objects
        source_tbl = self.source_db.table(source_table)
        destination_tbl = self.dest_db.table(destination_table)

        # Check that the destination table exists. If it does not, then run a
        # full sync instead.
        if not destination_tbl.exists:
            self.table_sync_full(source_table, destination_table)

        # If the source table contains 0 rows, do not attempt to copy the table.
        if source_tbl.num_rows == 0:
            logger.info('Source table contains 0 rows')
            return None

        # Check that the source table primary key is distinct
        if distinct_check and not source_tbl.distinct_primary_key(primary_key):
            raise ValueError('{primary_key} is not distinct in source table.')

        # Get the max source table and destination table primary key
        source_max_pk = source_tbl.max_primary_key(primary_key)
        dest_max_pk = destination_tbl.max_primary_key(primary_key)

        # Check for a mismatch in row counts; if dest_max_pk is None, or destination is empty
        # and we don't have to worry about this check.
        if dest_max_pk is not None and dest_max_pk > source_max_pk:
            raise ValueError('Destination DB primary key greater than source DB primary key.')

        # Do not copied if row counts are equal.
        elif dest_max_pk == source_max_pk:
            logger.info('Tables are in sync.')
            return None

        else:
            # Get count of rows to be copied.
            if dest_max_pk is not None:
                new_row_count = source_tbl.get_new_rows_count(primary_key, dest_max_pk)
            else:
                new_row_count = source_tbl.num_rows

            logger.info(f'Found {new_row_count} new rows in source table.')

            copied_rows = 0
            # Copy rows in chunks.
            while True:
                # Get a chunk
                rows = source_tbl.get_new_rows(primary_key=primary_key,
                                               cutoff_value=dest_max_pk,
                                               offset=copied_rows,
                                               chunk_size=self.chunk_size)

                row_count = rows.num_rows if rows else 0
                if row_count == 0:
                    break

                # Copy the chunk
                self.dest_db.copy(rows, destination_table, if_exists='append', **kwargs)

                # Update the counter
                copied_rows += row_count

        self._row_count_verify(source_tbl, destination_tbl)

        logger.info(f'{source_table} synced to {destination_table}.')

    def _row_count_verify(self, source_table_obj, destination_table_obj):
        """
        Ensure the the rows of the source table and the destination table match
        """

        source_row_count = source_table_obj.num_rows
        dest_row_count = destination_table_obj.num_rows

        if source_row_count != dest_row_count:
            logger.warning((f'Table count mismatch. Source table contains {source_row_count}.',
                           f' Destination table contains {dest_row_count}.'))
            return False

        logger.info('Source and destination table row counts match.')
        return True

    def _check_column_match(self, source_table_obj, destination_table_obj):
        """
        Ensure that the columns from each table match
        """

        if source_table_obj.columns != destination_table_obj.columns:
            raise ValueError("""Destination table columns do not match source table columns.
                             Consider dropping destination table and running a full sync.""")



from parsons.sftp.sftp import SFTP
from parsons.etl.table import Table
from parsons.utilities.files import create_temp_file
import xml.etree.ElementTree as ET
import os
import uuid
import time
import logging
import xmltodict


TS_STFP_HOST = 'sftp.targetsmart.com'
TS_SFTP_PORT = 2222
TS_SFTP_DIR = 'automation'

logger = logging.getLogger(__name__)

# Automation matching documentation can be found here:
# https://docs.targetsmart.com/developers/automation/index.html

# The columns are heavily customized by TS, so while I would like
# to do more column validation and mapping, I'm not sure that is
# going to be possible.


class TargetSmartAutomation(object):

    def __init__(self, sftp_username=None, sftp_password=None):

        self.sftp_host = TS_STFP_HOST
        self.sftp_port = TS_SFTP_PORT
        self.sftp_dir = TS_SFTP_DIR

        try:
            self.sftp_username = sftp_username or os.environ['TS_SFTP_USERNAME']
            self.sftp_password = sftp_password or os.environ['TS_SFTP_PASSWORD']

        except KeyError as error:
            logger.error("Connection info missing. It must include as kwarg or "
                         "env variable.")
            raise error

        self.sftp = SFTP(self.sftp_host, self.sftp_username, self.sftp_password,
                         self.sftp_port)

    def match(self, table, job_type, job_name=None, emails=None, call_back=None,
              remove_files=True):
        """
        Match a table to TargetSmart using their bulk matching service.

        .. warning::
            Table Columns
              The automation job does not validates the file by column indexes
              rather than columns names. So, if it expected 10 columns and you
              only provide 9, it will fail. However, if you provide 10 columns that
              are out of order, the job will succeed, but the records will not
              match.

        Args:
            table: Parsons Table Object
                A table object with the required columns. (Required columns provided be TargetSmart)
            job_type: str
                The match job type. **This is case sensitive.** (Match job names provided by TargetSmart)
            job_name: str
                Optional job name.
            emails: list
                A list of emails that will received status notifications. This
                is useful in debugging failed jobs.
            call_back: str
                A callback url to which the status will be posted. See
                `TargetSmart documentation <https://docs.targetsmart.com/developers/automation/index.html#http-callback>`_
                for more details.
            remove_files: boolean
                Remove the configuration, file to be matched and matched file from
                the TargetSmart FTP upon completion or failure of match.
        """ # noqa: E501,E261

        # Generate a match job
        job_name = job_name or str(uuid.uuid1())

        try:
            # Upload table
            self.sftp.put_file(table.to_csv(), f'{self.sftp_dir}/{job_name}_input.csv')
            logger.info(f'Table with {table.num_rows} rows uploaded to TargetSmart.')

            # Create/upload XML configuration
            xml = self.create_job_xml(job_type, job_name, emails=emails,
                                      status_key=job_name, call_back=call_back)
            self.sftp.put_file(xml, f'{self.sftp_dir}/{job_name}.job.xml')
            logger.info(f'Match configuration uploaded to TargetSmart.')

            # Check xml configuration status
            self.poll_config_status(job_name)

            # Check the status of the match
            self.match_status(job_name)

            # Download the resulting file
            tbl = Table.from_csv(self.sftp.get_file(f'{self.sftp_dir}/{job_name}_output.csv'))

        finally:
            # Clean up files
            if remove_files:
                self.remove_files(job_name)

            # Log Stats
            # TO DO: Provide some stats on the match

        # Return file as a Table
        return tbl

    def create_job_xml(self, job_type, job_name, emails=None, status_key=None, call_back=None):
        # Internal method to create a valid job xml

        job = ET.Element("job")

        # Generate Base XML
        input_file = ET.SubElement(job, 'inputfile')
        input_file.text = job_name + '_input.csv'
        output_file = ET.SubElement(job, 'outputfile')
        output_file.text = job_name + '_output.csv'
        jobtype = ET.SubElement(job, 'jobtype', text=job_type)
        jobtype.text = job_type

        # Add status key
        args = ET.SubElement(job, "args")
        statuskey = ET.SubElement(args, "arg", name="__status_key")
        statuskey.text = status_key or job_name

        # Option args
        if call_back:
            callback = ET.SubElement(args, "arg", name="__http_callback")
            callback.text = call_back

        if emails:
            emails_el = ET.SubElement(args, "arg", name="__emails")
            emails_el.text = ','.join(emails)

        # Write xml to file object
        local_path = create_temp_file(suffix='.xml')
        tree = ET.ElementTree(job)
        tree.write(local_path)
        return local_path

    def poll_config_status(self, job_name, polling_interval=20):
        #  Poll the configuration status

        while True:

            time.sleep(polling_interval)
            if self.config_status(job_name):
                return True
            logger.info(f'Waiting on {job_name} job configuration...')

    def config_status(self, job_name):
        # Check the status of the configuration by parsing the
        # the files in the SFTP directory.

        for f in self.sftp.list_directory(remote_path=self.sftp_dir):

            if f == f'{job_name}.job.xml.good':
                logger.info(f'Match job {job_name} configured.')
                return True

            elif f == f'{job_name}.job.xml.bad':
                logger.info(f'Match job {job_name} configuration error.')
                #  To Do: Lift up the configuration error.
                raise ValueError('Job configuration failed. If you provided an email'
                                 'address, you will be sent more details.')

            else:
                pass

        return False

    def match_status(self, job_name, polling_interval=60):
        # You could also poll their API for the status, which was what the original
        # version of the automation matching did. Note: The polling API is public
        # and does expose some metadata. This happens regardless of anything that
        # we do. However, the actually data is only exposed on the secure SFTP.

        while True:

            logger.debug('Match running...')
            for file_name in self.sftp.list_directory(remote_path=self.sftp_dir):

                if file_name == f'{job_name}.finish.xml':

                    xml_file = self.sftp.get_file(f'{self.sftp_dir}/{job_name}.finish.xml')
                    with open(xml_file, 'rb') as x:
                        xml = xmltodict.parse(x, dict_constructor=dict)

                    if xml['jobcontext']['state'] == 'error':
                        # To Do: Parse these in a pretty way
                        logger.info(f"Match Error: {xml['jobcontext']['errors']}")
                        raise ValueError(f"Match job failed. {xml['jobcontext']['errors']}")

                    elif xml['jobcontext']['state'] == 'success':
                        logger.info('Match complete.')

                        return True

            time.sleep(polling_interval)

    def remove_files(self, job_name):
        # Remove all of the files for the match.

        for file_name in self.sftp.list_directory(remote_path=self.sftp_dir):
            if job_name in file_name:
                self.sftp.remove_file(f'{self.sftp_dir}/{file_name}')
                logger.info(f'{file_name} removed from SFTP.')



import requests
from requests.auth import HTTPBasicAuth
from parsons.etl import Table
from parsons.utilities import check_env
import logging

logger = logging.getLogger(__name__)

PHONE2ACTION_URI = 'https://api.phone2action.com/2.0/'


class Phone2Action(object):
    """
    Instantiate Phone2Action Class

    `Args:`
        app_id: str
            The Phone2Action provided application id. Not required if ``PHONE2ACTION_APP_ID``
            env variable set.
        app_key: str
            The Phone2Action provided application key. Not required if ``PHONE2ACTION_APP_KEY``
            env variable set.
    `Returns:`
        Phone2Action Class
    """

    def __init__(self, app_id=None, app_key=None):

        self.app_id = check_env.check('PHONE2ACTION_APP_ID', app_id)
        self.app_key = check_env.check('PHONE2ACTION_APP_KEY', app_key)
        self.auth = HTTPBasicAuth(self.app_id, self.app_key)
        self.uri = PHONE2ACTION_URI

    def _request(self, url, args=None):
        # Internal request method

        r = requests.get(url, auth=self.auth, params=args)
        r.raise_for_status()
        return r

    def _paginate_request(self, url, args=None):
        # Internal pagination method

        r = self._request(url, args=args)

        json = r.json()['data']

        # If count of items is less than the total allowed per page, paginate
        while r.json()['pagination']['count'] == r.json()['pagination']['per_page']:

            r = self._request(r.json()['pagination']['next_url'], args)
            json.extend(r.json()['data'])

        return json

    def get_advocates(self, state=None, campaign_id=None, updated_since=None):
        """
        Return advocates (person records).

        `Args:`
            state: str
                Filter by US postal abbreviation for a state
                or territory e.g., "CA" "NY" or "DC"
            campaign_id: int
                Filter to specific campaign
            updated_since: str
                Fetch all advocates updated since UTC date time provided
                using (ex. '2014-01-05 23:59:43')
        `Returns:`
            A dict of parsons tables:
                * emails
                * phones
                * memberships
                * tags
                * ids
                * fields
                * advocates
        """

        url = self.uri + 'advocates'

        args = {'state': state,
                'campaignid': campaign_id,
                'updatedSince': updated_since}

        logger.info('Retrieving advocates...')
        json = self._paginate_request(url, args=args)

        return self._advocates_tables(Table(json))

    def _advocates_tables(self, tbl):
        # Convert the advocates nested table into multiple tables

        tbls = {}

        logger.info(f'Retrieved {tbl.num_rows} advocates...')

        # Unpack all of the single objects
        for c in ['created_at', 'updated_at', 'address', 'districts']:
            tbl.unpack_dict(c)

        # Unpack all of the arrays
        for c in ['emails', 'phones', 'memberships', 'tags', 'ids', 'fields']:
            tbls[c] = tbl.long_table(['id'], c, key_rename={'id': 'advocate_id'})

        # Add to tbls list
        tbls['advocates'] = tbl

        return tbls

    def get_campaigns(self, state=None, zip=None, include_generic=False, include_private=False,
                      include_content=True):
        """
        Returns a list of campaigns

        `Args:`
            state: str
                Filter by US postal abbreviation for a state or territory e.g., "CA" "NY" or "DC"
            zip: int
                Filter by 5 digit zip code
            include_generic: boolean
                When filtering by state or ZIP code, include unrestricted campaigns
            include_private: boolean
                If true, will include private campaigns in results
            include_content: boolean
                If true, include campaign content fields, which may vary. This may cause
                sync errors.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.uri + 'campaigns'

        args = {'state': state,
                'zip': zip,
                'includeGeneric': str(include_generic),
                'includePrivate': str(include_private)
                }

        tbl = Table(self._request(url, args=args).json())
        tbl.unpack_dict('updated_at')
        if include_content:
            tbl.unpack_dict('content')

        return tbl



test_campaigns = {
	"campaigns": [
		{
			"id": "abc",
			"web_id": 123,
			"type": "regular",
			"create_time": "2019-10-18T02:35:05+00:00",
			"archive_url": "http://example.com/sample-campaign-1",
			"long_archive_url": "https://mailchi.mp/abc/sample-campaign-1",
			"status": "sent",
			"emails_sent": 145,
			"send_time": "2019-10-18T14:15:00+00:00",
			"content_type": "template",
			"needs_block_refresh": False,
			"resendable": True,
			"recipients": {
				"list_id": "zyx",
				"list_is_active": True,
				"list_name": "Support Our Candidate List 1",
				"segment_text": "",
				"recipient_count": 145
			},
			"settings": {
				"subject_line": "Sample Campaign 1",
				"preview_text": "This is a sample campaign.",
				"title": "Sample Campaign Donation Ask",
				"from_name": "Our Candidate",
				"reply_to": "our_candidate@example.com",
				"use_conversation": False,
				"to_name": "*|FNAME|* *|LNAME|*",
				"folder_id": "",
				"authenticate": True,
				"auto_footer": False,
				"inline_css": False,
				"auto_tweet": False,
				"fb_comments": True,
				"timewarp": False,
				"template_id": 12345,
				"drag_and_drop": True
			},
			"tracking": {
				"opens": True,
				"html_clicks": True,
				"text_clicks": False,
				"goal_tracking": False,
				"ecomm360": False,
				"google_analytics": "",
				"clicktale": ""
			},
			"report_summary": {
				"opens": 48,
				"unique_opens": 34,
				"open_rate": 0.23776223776223776,
				"clicks": 1,
				"subscriber_clicks": 1,
				"click_rate": 0.006993006993006993,
				"ecommerce": {
					"total_orders": 0,
					"total_spent": 0,
					"total_revenue": 0
				}
			},
			"delivery_status": {
				"enabled": False
			},
			"_links": []
		},
		{
			"id": "def",
			"web_id": 456,
			"type": "regular",
			"create_time": "2019-05-29T11:46:41+00:00",
			"archive_url": "http://example.com/sample-campaign-2",
			"long_archive_url": "https://mailchi.mp/abc/sample-campaign-2",
			"status": "sent",
			"emails_sent": 87,
			"send_time": "2019-05-29T21:18:15+00:00",
			"content_type": "template",
			"needs_block_refresh": False,
			"resendable": True,
			"recipients": {
				"list_id": "wvu",
				"list_is_active": True,
				"list_name": "Support Our Candidate List 2",
				"segment_text": "",
				"recipient_count": 87
			},
			"settings": {
				"subject_line": "Sample Campaign 2",
				"preview_text": "This is another sample campaign.",
				"title": "Sample Campaign 2 Donation Ask",
				"from_name": "Our Candidate",
				"reply_to": "our_candidate@example.com",
				"use_conversation": False,
				"to_name": "*|FNAME|* *|LNAME|*",
				"folder_id": "",
				"authenticate": True,
				"auto_footer": False,
				"inline_css": False,
				"auto_tweet": False,
				"fb_comments": True,
				"timewarp": False,
				"template_id": 67890,
				"drag_and_drop": True
			},
			"tracking": {
				"opens": True,
				"html_clicks": True,
				"text_clicks": False,
				"goal_tracking": False,
				"ecomm360": False,
				"google_analytics": "",
				"clicktale": ""
			},
			"report_summary": {
				"opens": 108,
				"unique_opens": 48,
				"open_rate": 0.5647058823529412,
				"clicks": 25,
				"subscriber_clicks": 14,
				"click_rate": 0.16470588235294117,
				"ecommerce": {
					"total_orders": 0,
					"total_spent": 0,
					"total_revenue": 0
				}
			},
			"delivery_status": {
				"enabled": False
			},
			"_links": []
		}]}

test_lists = {
	"lists": [
		{
			"id": "zyx",
			"web_id": 98765,
			"name": "Support Our Candidate List 1",
			"contact": {
				"company": "Support Our Candidate",
				"address1": "123 Main Street",
				"address2": "",
				"city": "Townsville",
				"state": "OH",
				"zip": "43358",
				"country": "US",
				"phone": ""
			},
			"permission_reminder": "You are receiving this email because you signed up at an event, while being canvassed, or on our website.",
			"use_archive_bar": True,
			"campaign_defaults": {
				"from_name": "Our Candidate",
				"from_email": "our_candidate@example.com",
				"subject": "",
				"language": "en"
			},
			"notify_on_subscribe": "",
			"notify_on_unsubscribe": "",
			"date_created": "2019-03-25T22:55:44+00:00",
			"list_rating": 3,
			"email_type_option": False,
			"subscribe_url_short": "http://example.com/sample-subscribe_url_2",
			"subscribe_url_long": "https://mailchi.mp/zyx/sample-subscribe-url-2",
      "beamer_address": "us00-sample-2@inbound.mailchimp.com",
			"visibility": "pub",
			"double_optin": False,
			"has_welcome": True,
			"marketing_permissions": False,
			"modules": [],
			"stats": {
				"member_count": 140,
				"unsubscribe_count": 8,
				"cleaned_count": 16,
				"member_count_since_send": 0,
				"unsubscribe_count_since_send": 1,
				"cleaned_count_since_send": 0,
				"campaign_count": 21,
				"campaign_last_sent": "2020-01-06T01:54:32+00:00",
				"merge_field_count": 5,
				"avg_sub_rate": 0,
				"avg_unsub_rate": 1,
				"target_sub_rate": 3,
				"open_rate": 38.40236686390532,
				"click_rate": 4.016786570743405,
				"last_sub_date": "2019-09-24T01:07:56+00:00",
				"last_unsub_date": "2020-01-06T01:55:02+00:00"
			},
			"_links": []
		},
    {
			"id": "xvu",
			"web_id": 43210,
			"name": "Support Our Candidate List 2",
			"contact": {
				"company": "Support Our Candidate",
				"address1": "123 Main Street",
				"address2": "",
				"city": "Townsville",
				"state": "OH",
				"zip": "43358",
				"country": "US",
				"phone": ""
			},
			"permission_reminder": "You are receiving this email because you signed up at an event, while being canvassed, or on our website.",
			"use_archive_bar": True,
			"campaign_defaults": {
				"from_name": "Our Candidate",
				"from_email": "our_candidate@example.com",
				"subject": "",
				"language": "en"
			},
			"notify_on_subscribe": "",
			"notify_on_unsubscribe": "",
			"date_created": "2018-09-15T22:15:21+00:00",
			"list_rating": 3,
			"email_type_option": False,
			"subscribe_url_short": "http://example.com/sample-subscribe_url_1",
			"subscribe_url_long": "https://mailchi.mp/zyx/sample-subscribe-url-1",
      "beamer_address": "us00-sample-1@inbound.mailchimp.com",
			"visibility": "pub",
			"double_optin": False,
			"has_welcome": True,
			"marketing_permissions": False,
			"modules": [],
			"stats": {
				"member_count": 73,
				"unsubscribe_count": 3,
				"cleaned_count": 7,
				"member_count_since_send": 1,
				"unsubscribe_count_since_send": 1,
				"cleaned_count_since_send": 0,
				"campaign_count": 13,
				"campaign_last_sent": "2020-01-03T14:38:11+00:00",
				"merge_field_count": 5,
				"avg_sub_rate": 0,
				"avg_unsub_rate": 1,
				"target_sub_rate": 3,
				"open_rate": 64.19236186394533,
				"click_rate": 3.746759370417411,
				"last_sub_date": "2020-01-01T00:19:46+00:00",
				"last_unsub_date": "2019-12-23T11:44:31+00:00"
			},
			"_links": []
		},
	],
	"total_items": 1,
	"constraints": {
		"may_create": False,
		"max_instances": 1,
		"current_total_instances": 1
	},
	"_links": []
}

test_members = {
  "members": [
    {
      "id": "9eb69db8d0371811aa18803a1ae21584",
      "email_address": "member_1@example.com",
      "unique_email_id": "c82a25d939",
      "web_id": 24816326,
      "email_type": "html",
      "status": "subscribed",
      "merge_fields": {
        "FNAME": "Member",
        "LNAME": "One",
        "ADDRESS": {
          "addr1": "",
          "addr2": "",
          "city": "",
          "state": "",
          "zip": "",
          "country": "US"
        },
        "PHONE": "",
        "BIRTHDAY": ""
      },
      "stats": {
        "avg_open_rate": 0.3571,
        "avg_click_rate": 0
      },
      "ip_signup": "",
      "timestamp_signup": "",
      "ip_opt": "174.59.50.35",
      "timestamp_opt": "2019-03-25T22:55:44+00:00",
      "member_rating": 4,
      "last_changed": "2019-03-25T22:55:44+00:00",
      "language": "en",
      "vip": False,
      "email_client": "Gmail",
      "location": {
        "latitude": 40.0293,
        "longitude": -76.2656,
        "gmtoff": 0,
        "dstoff": 0,
        "country_code": "US",
        "timezone": "717/223"
      },
      "source": "Unknown",
      "tags_count": 0,
      "tags": [],
      "list_id": "67fdf4b1f4",
      "_links": []
    },
    {
      "id": "4f315641dbad7b74acc0f4a5d3741ac6",
      "email_address": "member_2@example.com",
      "unique_email_id": "8d308d69d3",
      "web_id": 12233445,
      "email_type": "html",
      "status": "subscribed",
      "merge_fields": {
        "FNAME": "Member",
        "LNAME": "Two",
        "ADDRESS": "",
        "PHONE": "",
        "BIRTHDAY": ""
      },
      "stats": {
        "avg_open_rate": 0.5,
        "avg_click_rate": 0
      },
      "ip_signup": "",
      "timestamp_signup": "",
      "ip_opt": "174.59.50.35",
      "timestamp_opt": "2019-03-25T23:04:46+00:00",
      "member_rating": 4,
      "last_changed": "2019-03-25T23:04:46+00:00",
      "language": "",
      "vip": False,
      "email_client": "iPhone",
      "location": {
        "latitude": 40.0459,
        "longitude": -76.3542,
        "gmtoff": 0,
        "dstoff": 0,
        "country_code": "US",
        "timezone": "717/223"
      },
      "source": "Import",
      "tags_count": 2,
      "tags": [
        {
          "id": 17493,
          "name": "canvass"
        },
        {
          "id": 17497,
          "name": "canvass-03-17-2019"
        }
      ],
      "list_id": "67fdf4b1f4",
      "_links": []
    }]}

test_unsubscribes = {
  "unsubscribes": [
    {
      "email_id": "e542e5cd7b414e5ff8409ff57cf154be",
      "email_address": "unsubscribe_1@exmaple.com",
      "merge_fields": {
        "FNAME": "Unsubscriber",
        "LNAME": "One",
        "ADDRESS": "",
        "PHONE": "5558754307",
        "BIRTHDAY": ""
      },
      "vip": False,
      "timestamp": "2019-12-09T21:18:06+00:00",
      "reason": "None given",
      "campaign_id": "abc",
      "list_id": "zyx",
      "list_is_active": True,
      "_links": []
    }
  ],
  "campaign_id": "abc",
  "total_items": 1,
  "_links": []
}


from parsons.utilities import json_format
import logging

logger = logging.getLogger(__name__)


class People(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def find_person(self, first_name=None, last_name=None, date_of_birth=None, email=None,
                    phone=None, phone_type=None, street_number=None, street_name=None, zip=None):
        """
        Find a person record.

        .. note::
            Person find must include the following minimum combinations to conduct
            a search.

            - first_name, last_name, email
            - first_name, last_name, phone
            - first_name, last_name, zip5, date_of_birth
            - first_name, last_name, street_number, street_name, zip5
            - email_address

        `Args:`
            first_name: str
                The person's first name
            last_name: str
                The person's last name
            dob: str
                ISO 8601 formatted date of birth (e.g. ``1981-02-01``)
            email: str
                The person's email address
            phone: str
                Phone number of any type (Work, Cell, Home)
            street_number: str
                Street Number
            street_name: str
                Street Name
            zip: str
                5 digit zip code
        `Returns:`
            A person dict object
        """

        logger.info(f'Finding {first_name} {last_name}.')

        return self._people_search(
            first_name=first_name,
            last_name=last_name,
            date_of_birth=date_of_birth,
            email=email,
            phone=phone,
            phone_type=phone_type,
            street_number=street_number,
            street_name=street_name,
            zip=zip
        )

    def find_person_json(self, match_json):
        """
        Find a person record based on json data.

        .. note::
            Person find must include the following minimum combinations to conduct
            a search.

            - first_name, last_name, email
            - first_name, last_name, phone
            - first_name, last_name, zip5, date_of_birth
            - first_name, last_name, street_number, street_name, zip5
            - email_address

        .. note::
            A full list of possible values for the json, and its structure can be found
            `here <https://developers.ngpvan.com/van-api#match-candidates>`_.

        `Args:`
            match_json: dict
                A dictionary of values to match against.
            fields: The fields to return. Leave as default for all available fields
        `Returns:`
            A person dict object
        """

        logger.info(f'Finding a match for json details.')

        return self._people_search(match_json=match_json)

    def update_person(self, id=None, id_type='vanid', first_name=None, last_name=None,
                      date_of_birth=None, email=None, phone=None, phone_type=None,
                      street_number=None, street_name=None, zip=None):
        """
        Update a person record based on a provided ID. All other arguments provided will be
        updated on the record.

        .. warning::
            This method can only be run on MyMembers, EveryAction, MyCampaign databases.

        `Args:`
            id: str
                A valid id
            id_type: str
                A known person identifier type available on this VAN instance.
                Defaults to ``vanid``.
            first_name: str
                The person's first name
            last_name: str
                The person's last name
            dob: str
                ISO 8601 formatted date of birth (e.g. ``1981-02-01``)
            email: str
                The person's email address
            phone: str
                Phone number of any type (Work, Cell, Home)
            phone_type: str
                One of 'H' for home phone, 'W' for work phone, 'C' for cell, 'M' for
                main phone or 'F' for fax line. Defaults to home phone.
            street_number: str
                Street Number
            street_name: str
                Street Name
            zip: str
                5 digit zip code
        `Returns:`
            A person dict
        """

        return self._people_search(
            id=id,
            id_type=id_type,
            first_name=first_name,
            last_name=last_name,
            date_of_birth=date_of_birth,
            email=email,
            phone=phone,
            phone_type=phone_type,
            street_number=street_number,
            street_name=street_name,
            zip=zip,
            create=True
        )

    def update_person_json(self, id, id_type='vanid', match_json=None):
        """
        Update a person record based on a provided ID within the match_json dict.

        .. note::
            A full list of possible values for the json, and its structure can be found
            `here <https://developers.ngpvan.com/van-api#match-candidates>`_.

        `Args:`
            id: str
                A valid id
            id_type: str
                A known person identifier type available on this VAN instance.
                Defaults to ``vanid``.
            match_json: dict
                A dictionary of values to match against and save.
        `Returns:`
            A person dict
        """

        return self._people_search(id=id, id_type=id_type, match_json=match_json, create=True)

    def upsert_person(self, first_name=None, last_name=None, date_of_birth=None, email=None,
                      phone=None, phone_type=None, street_number=None, street_name=None, zip=None):
        """
        Create or update a person record.

        .. note::
            Person find must include the following minimum combinations.
            - first_name, last_name, email
            - first_name, last_name, phone
            - first_name, last_name, zip5, date_of_birth
            - first_name, last_name, street_number, street_name, zip5
            - email_address

        .. warning::
            This method can only be run on MyMembers, EveryAction, MyCampaign databases.

        `Args:`
            first_name: str
                The person's first name
            last_name: str
                The person's last name
            dob: str
                ISO 8601 formatted date of birth (e.g. ``1981-02-01``)
            email: str
                The person's email address
            phone: str
                Phone number of any type (Work, Cell, Home)
            phone_type: str
                One of 'H' for home phone, 'W' for work phone, 'C' for cell, 'M' for
                main phone or 'F' for fax line. Defaults to home phone.
            street_number: str
                Street Number
            street_name: str
                Street Name
            zip: str
                5 digit zip code
        `Returns:`
            A person dict
        """

        return self._people_search(
            first_name=first_name,
            last_name=last_name,
            date_of_birth=date_of_birth,
            email=email,
            phone=phone,
            phone_type=phone_type,
            street_number=street_number,
            street_name=street_name,
            zip=zip,
            create=True
        )

    def upsert_person_json(self, match_json):
        """
        Create or update a person record.

        .. note::
            Person find must include the following minimum combinations.
            - first_name, last_name, email
            - first_name, last_name, phone
            - first_name, last_name, zip5, date_of_birth
            - first_name, last_name, street_number, street_name, zip5
            - email_address

        .. note::
            A full list of possible values for the json, and its structure can be found
            `here <https://developers.ngpvan.com/van-api#match-candidates>`_. `vanId` can
            be passed to ensure the correct record is updated.

        .. warning::
            This method can only be run on MyMembers, EveryAction, MyCampaign databases.

        `Args:`
            match_json: dict
                A dictionary of values to match against and save.
        `Returns:`
            A person dict
        """

        return self._people_search(match_json=match_json, create=True)

    def _people_search(self, id=None, id_type=None, first_name=None, last_name=None,
                       date_of_birth=None, email=None, phone=None, phone_type='H',
                       street_number=None, street_name=None, zip=None, match_json=None,
                       create=False):
        # Internal method to hit the people find/create endpoints

        addressLine1 = None
        if street_name and street_number:
            addressLine1 = f'{street_number} {street_name}'

        # Check to see if a match map has been provided
        if not match_json:
            json = {"firstName": first_name, "lastName": last_name}

            # Will fail if empty dicts are provided, hence needed to add if exist
            if email:
                json['emails'] = [{'email': email}]
            if phone:  # To Do: Strip out non-integers from phone
                json['phones'] = [{'phoneNumber': phone, 'phoneType': phone_type}]
            if date_of_birth:
                json['dateOfBirth'] = date_of_birth
            if zip or addressLine1:
                json['addresses'] = [{}]
                if zip:
                    json['addresses'][0]['zipOrPostalCode'] = zip
                if addressLine1:
                    json['addresses'][0]['addressLine1'] = addressLine1
        else:
            json = match_json
            if 'vanId' in match_json:
                id = match_json['vanId']

        url = 'people/'

        if id:

            if create:
                id_type = '' if id_type in ('vanid', None) else f"{id_type}:"
                url += id_type + str(id)
            else:
                return self.get_person(id, id_type=id_type)

        else:
            url += 'find'

            if create:
                url += 'OrCreate'
            else:
                # Ensure that the minimum combination of fields were passed
                json_flat = json_format.flatten_json(json)
                self._valid_search(**json_flat)

        return self.connection.post_request(url, json=json)

    def _valid_search(self, firstName=None, lastName=None, email=None, phoneNumber=None,
                      dateOfBirth=None, addressLine1=None, zipOrPostalCode=None, **kwargs):
        # Internal method to check if a search is valid, kwargs are ignored

        if (None in [firstName, lastName, email] and
            None in [firstName, lastName, phoneNumber] and
            None in [firstName, lastName, zipOrPostalCode, dateOfBirth] and
            None in [firstName, lastName, addressLine1, zipOrPostalCode] and
                None in [email]):

            raise ValueError("""
                             Person find must include the following minimum
                             combinations to conduct a search.
                                - first_name, last_name, email
                                - first_name, last_name, phone
                                - first_name, last_name, zip, dob
                                - first_name, last_name, street_number, street_name, zip
                                - email
                            """)

        return True

    def get_person(self, id, id_type='vanid', expand_fields=[
                   'contribution_history', 'addresses', 'phones', 'emails',
                   'codes', 'custom_fields', 'external_ids', 'preferences',
                   'recorded_addresses', 'reported_demographics', 'suppressions',
                   'cases', 'custom_properties', 'districts', 'election_records',
                   'membership_statuses', 'notes', 'organization_roles',
                   'disclosure_field_values']):
        """
        Returns a single person record using their VANID or external id.

        `Args:`
            id: str
                A valid id
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``. Defaults to ``vanid``.
            expand_fields: list
                A list of fields for which to include data. If a field is omitted,
                ``None`` will be returned for that field. Can be ``contribution_history``,
                ``addresses``, ``phones``, ``emails``, ``codes``, ``custom_fields``,
                ``external_ids``, ``preferences``, ``recorded_addresses``,
                ``reported_demographics``, ``suppressions``, ``cases``, ``custom_properties``,
                ``districts``, ``election_records``, ``membership_statuses``, ``notes``,
                ``organization_roles``, ``scores``, ``disclosure_field_values``.
        `Returns:`
            A person dict
        """

        # Change end point based on id type
        url = 'people/'

        id_type = '' if id_type in ('vanid', None) else f"{id_type}:"
        url += id_type + str(id)

        expand_fields = ','.join([json_format.arg_format(f) for f in expand_fields])

        logger.info(f'Getting person with {id_type} of {id} at url {url}')
        return self.connection.get_request(url, params={'$expand': expand_fields})

    def apply_canvass_result(self, id, result_code_id, id_type='vanid', contact_type_id=None,
                             input_type_id=None, date_canvassed=None):
        """
        Apply a canvass result to a person. Use this end point for attempts that do not
        result in a survey response or an activist code (e.g. Not Home).

        `Args:`
            id: str
                A valid person id
            result_code_id : int
                Specifies the result code of the attempt. Valid ids can be found
                by using the :meth:`get_canvass_responses_result_codes`
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            contact_type_id : int
                `Optional`; A valid contact type id
            input_type_id : int
                `Optional`; Defaults to 11 (API Input)
            date_canvassed : str
                `Optional`; ISO 8601 formatted date. Defaults to todays date
        `Returns:`
            ``None``
        """

        logger.info(f'Applying result code {result_code_id} to {id_type} {id}.')
        self.apply_response(id, None, id_type=id_type, contact_type_id=contact_type_id,
                            input_type_id=input_type_id, date_canvassed=date_canvassed,
                            result_code_id=result_code_id)

    def toggle_volunteer_action(self, id, volunteer_activity_id, action, id_type='vanid',
                                result_code_id=None, contact_type_id=None, input_type_id=None,
                                date_canvassed=None):
        """
        Apply or remove a volunteer action to or from a person.

        `Args:`
            id: str
                A valid person id
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            volunteer_activity_id: int
                A valid volunteer activity id
            action: str
                Either 'apply' or 'remove'
            result_code_id : int
                `Optional`; Specifies the result code of the response. If
                not included,responses must be specified. Conversely, if
                responses are specified, result_code_id must be null. Valid ids
                can be found by using the :meth:`get_canvass_responses_result_codes`
            contact_type_id: int
                `Optional`; A valid contact type id
            input_type_id: int
                `Optional`; Defaults to 11 (API Input)
            date_canvassed: str
                `Optional`; ISO 8601 formatted date. Defaults to todays date

        ** NOT IMPLEMENTED **
        """

        """
        response = {"volunteerActivityId": volunteer_activity_id,
                    "action": self._action_parse(action),
                    "type": "VolunteerActivity"}

        logger.info(f'{action} volunteer activity {volunteer_activity_id} to {id_type} {id}')
        self.apply_response(id, response, id_type, contact_type_id, input_type_id, date_canvassed,
                            result_code_id)
        """

    def apply_response(self, id, response, id_type='vanid', contact_type_id=None,
                       input_type_id=None, date_canvassed=None, result_code_id=None):
        """
        Apply responses such as survey questions, activist codes, and volunteer actions
        to a person record. This method allows you apply multiple responses (e.g. two survey
        questions) at the same time. It is a low level method that requires that you
        conform to the VAN API `response object format <https://developers.ngpvan.com/van-api#people-post-people--vanid--canvassresponses>`_.

        `Args:`
            id: str
                A valid person id
            response: dict
                A list of dicts with each dict containing a valid action.
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            result_code_id : int
                `Optional`; Specifies the result code of the response. If
                not included,responses must be specified. Conversely, if
                responses are specified, result_code_id must be null. Valid ids
                can be found by using the :meth:`get_canvass_responses_result_codes`
            contact_type_id : int
                `Optional`; A valid contact type id
            input_type_id : int
                `Optional`; Defaults to 11 (API Input)
            date_canvassed : str
                `Optional`; ISO 8601 formatted date. Defaults to todays date
            responses : list or dict
        `Returns:`
            ``True`` if successful

        .. code-block:: python

            response = [{"activistCodeId": 18917,
                         "action": "Apply",
                         "type": "ActivistCode"},
                        {"surveyQuestionId": 109149,
                         "surveyResponseId": 465468,
                         "action": "SurveyResponse"}
                        ]
            van.apply_response(5222, response)
        """  # noqa: E501,E261

        # Set url based on id_type
        if id_type == 'vanid':
            url = f"people/{id}/canvassResponses"
        else:
            url = f"people/{id_type}:{id}/canvassResponses"

        json = {"canvassContext": {
            "contactTypeId": contact_type_id,
            "inputTypeId": input_type_id,
            "dateCanvassed": date_canvassed},
            "resultCodeId": result_code_id}

        if response:
            json['responses'] = response

        if result_code_id is not None and response is not None:
            raise ValueError("Both result_code_id and responses cannot be specified.")

        if isinstance(response, dict):
            json["responses"] = [response]

        if result_code_id is not None and response is not None:
            raise ValueError(
                "Both result_code_id and responses cannot be specified.")

        return self.connection.post_request(url, json=json)

    def create_relationship(self, vanid_1, vanid_2, relationship_id):
        """
        Create a relationship between two individuals

        `Args:`
            vanid_1 : int
                The vanid of the primary individual; aka the node
            vanid_2 : int
                The vanid of the secondary individual; the spoke
            relationship_id : int
                The relationship id indicating the type of relationship
        `Returns:`
            ``None``
        """

        json = {'relationshipId': relationship_id,
                'vanId': vanid_2}

        self.connection.post_request(f"people/{vanid_1}/relationships", json=json)
        logger.info('Relationship {vanid_1} to {vanid_2} created.')

    def apply_person_code(self, id, code_id, id_type='vanid'):
        """
        Apply a code to a person.

        `Args:`
            id: str
                A valid person id.
            code_id: int
                A valid code id.
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
        `Returns:`
            ``None``
        """

        # Set url based on id_type
        if id_type == 'vanid':
            url = f"people/{id}/codes"
        else:
            url = f"people/{id_type}:{id}/codes"

        json = {"codeId": code_id}

        self.connection.post_request(url, json=json)
        logger.info(f'Code {code_id} applied to person id {id}.')



from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector
from parsons import Table
import logging
import jwt
import datetime

logger = logging.getLogger(__name__)

ZOOM_URI = 'https://api.zoom.us/v2/'


class Zoom:
    """
    Instantiate the Zoom class.

    `Args:`
        api_key: str
            A valid Zoom api key. Not required if ``ZOOM_API_KEY`` env
            variable set.
        api_secret:
            A valid Zoom api secret. Not required if ``ZOOM_API_SECRET`` env
            variable set.
    """

    def __init__(self, api_key=None, api_secret=None):

        self.api_key = check_env.check('ZOOM_API_KEY', api_key)
        self.api_secret = check_env.check('ZOOM_API_SECRET', api_secret)
        self.client = APIConnector(ZOOM_URI)

    def refresh_header_token(self):
        # Generate a token that is valid for 30 seconds and update header. Full documentation
        # on JWT generation using Zoom API: https://marketplace.zoom.us/docs/guides/auth/jwt

        payload = {"iss": self.api_key, "exp": int(datetime.datetime.now().timestamp() + 30)}
        token = jwt.encode(payload, self.api_secret, algorithm='HS256').decode("utf-8")
        self.client.headers = {'authorization': f"Bearer {token}",
                               'content-type': "application/json"}

    def get_request(self, endpoint, data_key, params=None, **kwargs):
        # Internal GET request method.

        # To Do: Consider increasing default page size.

        self.refresh_header_token()
        r = self.client.get_request(ZOOM_URI + endpoint, params=params, **kwargs)
        self.client.data_key = data_key
        data = self.client.data_parse(r)

        if not params:
            params = {}

        # Return a dict or table if only one item.
        if 'page_number' not in r.keys():
            if isinstance(data, dict):
                return data
            if isinstance(data, list):
                return Table(data)

        # Else iterate through the pages and return a Table
        else:
            while r['page_number'] < r['page_count']:
                params['page_number'] = int(r['page_number']) + 1
                r = self.client.get_request(ZOOM_URI + endpoint, params=params, **kwargs)
                data.extend(self.client.data_parse(r))
            return Table(data)

    def get_users(self, status='active', role_id=None):
        """
        Get users.

        `Args:`
            status: str
                Filter by the user status. Must be one of following: ``active``,
                ``inactive``, or ``pending``.
            role_id: str
                Filter by the user role.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if status not in ['active', 'inactive', 'pending']:
            raise ValueError('Invalid status type provided.')

        params = {'status': status,
                  'role_id': role_id}

        tbl = self.get_request('users', 'users', params=params)
        logger.info(f'Retrieved {tbl.num_rows} users.')
        return tbl

    def get_meetings(self, user_id, meeting_type='scheduled'):
        """
        Get meetings scheduled by a user.

        `Args:`
            user_id: str
                A user id or email address of the meeting host.
            meeting_type: str

                .. list-table::
                    :widths: 25 50
                    :header-rows: 1

                    * - Type
                      - Notes
                    * - ``scheduled``
                      - This includes all valid past meetings, live meetings and upcoming
                        scheduled meetings. It is the equivalent to the combined list of
                        "Previous Meetings" and "Upcoming Meetings" displayed in the user's
                        Meetings page.
                    * - ``live``
                      - All the ongoing meetings.
                    * - ``upcoming``
                      - All upcoming meetings including live meetings.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self.get_request(f'users/{user_id}/meetings', 'meetings')
        logger.info(f'Retrieved {tbl.num_rows} meetings.')
        return tbl

    def get_past_meeting(self, meeting_uuid):
        """
        Get metadata regarding a past meeting.

        `Args:`
            meeting_id: str
                The meeting id
        `Returns:`
            dict
        """

        tbl = self.get_request(f'past_meetings/{meeting_uuid}', None)
        logger.info(f'Retrieved meeting {meeting_uuid}.')
        return tbl

    def get_past_meeting_participants(self, meeting_id):
        """
        Get past meeting participants

        `Args:`
            meeting_id:
                The meeting id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self.get_request(f'/report/meetings/{meeting_id}/participants', 'participants')
        logger.info(f'Retrieved {tbl.num_rows} participants.')
        return tbl



import unittest
import os
import requests_mock
from parsons.mobilize_america import MobilizeAmerica
from test.utils import validate_list


class TestMobilizeAmerica(unittest.TestCase):

    def setUp(self):

        self.ma = MobilizeAmerica()

    def tearDown(self):

        pass

    """
    # Does not work locally due to some UTC issues, but works on CircleCI. Commenting
    # out for the time being.

    def test_time_parse(self):

        # Test that Unix conversion works correctly
        self.assertEqual(self.ma._time_parse('<=2018-12-13'), 'lte_1544659200')

        # Test that it throws an error when you put in an invalid filter
        self.assertRaises(ValueError, self.ma._time_parse, '=2018-12-01')
    """

    @requests_mock.Mocker()
    def test_get_organizations(self, m):

        json = {
            "count": 38,
            "next": None,
            "previous": "https://events.mobilizeamerica.io/api/v1/organizations?updated_since=1543644000",
            "data": [
                        {
                            "id": 1251,
                            "name": "Mike Blake for New York City",
                            "slug": "mikefornyc",
                            "is_coordinated": 'True',
                            "is_independent": 'True',
                            "is_primary_campaign": 'False',
                            "state": "",
                            "district": "",
                            "candidate_name": "",
                            "race_type": "OTHER_LOCAL",
                            "event_feed_url": "https://events.mobilizeamerica.io/mikefornyc/",
                            "created_date": 1545885434,
                            "modified_date": 1546132256
                        }
            ]
        }

        m.get(self.ma.uri + 'organizations', json=json)

        expected = ['id', 'name', 'slug', 'is_coordinated', 'is_independent', 'is_primary_campaign',
                    'state', 'district', 'candidate_name', 'race_type', 'event_feed_url', 'created_date',
                    'modified_date']

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.ma.get_organizations()))

    @requests_mock.Mocker()
    def test_get_events(self, m):

        json = {'count': 1, 'next': None, 'previous': None, 'data': [{'id': 86738, 'description': 'Join our team of volunteers and learn how to engage students in local high schools, communicate our mission, and register young voters.', 'timezone': 'America/Chicago', 'title': 'Student Voter Initiative Training', 'summary': '', 'featured_image_url': 'https://mobilizeamerica.imgix.net/uploads/event/40667432145_6188839fe3_o_20190102224312253645.jpeg', 'sponsor': {'id': 1076, 'name': 'Battleground Texas', 'slug': 'battlegroundtexas', 'is_coordinated': True, 'is_independent': False, 'is_primary_campaign': False, 'state': '', 'district': '', 'candidate_name': '', 'race_type': None, 'event_feed_url': 'https://events.mobilizeamerica.io/battlegroundtexas/',
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'created_date': 1538590930, 'modified_date': 1546468308}, 'timeslots': [{'id': 526226, 'start_date': 1547330400, 'end_date': 1547335800}], 'location': {'venue': 'Harris County Democratic Party HQ', 'address_lines': ['4619 Lyons Ave', ''], 'locality': 'Houston', 'region': 'TX', 'postal_code': '77020', 'location': {'latitude': 29.776446, 'longitude': -95.323037}, 'congressional_district': '18', 'state_leg_district': '142', 'state_senate_district': None}, 'event_type': 'TRAINING', 'created_date': 1546469706, 'modified_date': 1547335800, 'browser_url': 'https://events.mobilizeamerica.io/battlegroundtexas/event/86738/', 'high_priority': None, 'contact': None, 'visibility': 'PUBLIC'}]}
        m.get(self.ma.uri + 'events', json=json)

        expected = ['id', 'description', 'timezone', 'title', 'summary', 'featured_image_url', 'event_type', 'created_date', 'modified_date', 'browser_url', 'high_priority', 'contact', 'visibility', 'sponsor_candidate_name', 'sponsor_created_date', 'sponsor_district', 'sponsor_event_feed_url', 'sponsor_id', 'sponsor_is_coordinated', 'sponsor_is_independent',
                    'sponsor_is_primary_campaign', 'sponsor_modified_date', 'sponsor_name', 'sponsor_race_type', 'sponsor_slug', 'sponsor_state', 'address_lines', 'congressional_district', 'locality', 'postal_code', 'region', 'state_leg_district', 'state_senate_district', 'venue', 'latitude', 'longitude', 'timeslots_0_end_date', 'timeslots_0_id', 'timeslots_0_start_date']

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.ma.get_events()))

    @requests_mock.Mocker()
    def test_get_events_deleted(self, m):

        json = {'count': 2,
                'next': None,
                'previous': None,
                'data': [{'id': 86765,
                          'deleted_date': 1546705971},
                         {'id': 86782,
                          'deleted_date': 1546912779}
                         ]
                }

        m.get(self.ma.uri + 'events/deleted', json=json)

        # Assert response is expected structure
        self.assertTrue(validate_list(['id', 'deleted_date'], self.ma.get_events_deleted()))



from parsons.mailchimp.mailchimp import Mailchimp
import unittest
import requests_mock
from test.utils import assert_matching_tables
from test.test_mailchimp import expected_json

API_KEY = 'mykey-us00'

class TestMailchimp(unittest.TestCase):

    def setUp(self):

        self.mc = Mailchimp(API_KEY)

    @requests_mock.Mocker()
    def test_get_campaigns(self, m):

    	# Test that campaigns are returned correctly.
        m.get(self.mc.uri + 'campaigns', json=expected_json.test_campaigns)
        tbl = self.mc.get_campaigns()

        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_get_lists(self, m):

        # Test that lists are returned correctly.
        m.get(self.mc.uri + 'lists', json=expected_json.test_lists)
        tbl = self.mc.get_lists()

        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_get_members(self, m):

        # Test that list members are returned correctly.
        m.get(self.mc.uri + 'lists/zyx/members', json=expected_json.test_members)
        tbl = self.mc.get_members(list_id='zyx') 

        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_get_unsubscribes(self, m):

        # Test that campaign unsubscribes are returned correctly.
        m.get(self.mc.uri + 'reports/abc/unsubscribed', json=expected_json.test_unsubscribes)
        tbl = self.mc.get_unsubscribes(campaign_id='abc') 

        self.assertEqual(tbl.num_rows, 1)



from parsons.databases.postgres.postgres_core import PostgresCore
from parsons.databases.table import BaseTable
import logging
import os


logger = logging.getLogger(__name__)


class Postgres(PostgresCore):
    """
    A Postgres class to connect to database. Credentials can be passed from a ``.pgpass`` file
    stored in your home directory or with environmental variables.

    Args:
        username: str
            Required if env variable ``PGUSER`` not populated
        password: str
            Required if env variable ``PGPASSWORD`` not populated
        host: str
            Required if env variable ``PGHOST`` not populated
        db: str
            Required if env variable ``PGDATABASE`` not populated
        port: int
            Required if env variable ``PGPORT`` not populated.
        timeout: int
            Seconds to timeout if connection not established.
    """

    def __init__(self, username=None, password=None, host=None, db=None, port=5432, timeout=10):

        self.username = username or os.environ.get('PGUSER')
        self.password = password or os.environ.get('PGPASSWORD')
        self.host = host or os.environ.get('PGHOST')
        self.db = db or os.environ.get('PGDATABASE')
        self.port = port or os.environ.get('PGPORT')

        # Check if there is a pgpass file. Psycopg2 will search for this file first when
        # creating a connection.
        pgpass = os.path.isfile(os.path.expanduser('~/.pgpass'))

        if not any([self.username, self.password, self.host, self.db]) and not pgpass:
            raise ValueError('Connection arguments missing. Please pass as a pgpass file, kwargs',
                             'or env variables.')

        self.timeout = timeout
        self.dialect = 'postgres'

    def copy(self, tbl, table_name, if_exists='fail'):
        """
        Copy a :ref:`parsons-table` to Postgres.

        tbl: parsons.Table
            A Parsons table object
        table_name: str
            The destination schema and table (e.g. ``my_schema.my_table``)
        if_exists: str
            If the table already exists, either ``fail``, ``append``, ``drop``
            or ``truncate`` the table.
        """

        with self.connection() as connection:

            # Auto-generate table
            if self._create_table_precheck(connection, table_name, if_exists):

                # Create the table
                # To Do: Pass in the advanced configuration parameters.
                sql = self.create_statement(tbl, table_name)

                self.query_with_connection(sql, connection, commit=False)
                logger.info(f'{table_name} created.')

            sql = f"COPY {table_name} FROM STDIN CSV HEADER;"

            with self.cursor(connection) as cursor:
                cursor.copy_expert(sql, open(tbl.to_csv(), "r"))
                logger.info(f'{tbl.num_rows} rows copied to {table_name}.')

    def table(self, table_name):
        # Return a Postgres table object

        return PostgresTable(self, table_name)


class PostgresTable(BaseTable):
    # Postgres table object.

    pass



from parsons.etl import Table
import requests
import logging
from parsons.utilities import check_env

logger = logging.getLogger(__name__)

TURBOVOTE_URI = 'https://turbovote-admin-http-api.prod.democracy.works/'


class TurboVote(object):
    """
    Initialize Turbovote class

    `Args:`
        username: str
            A valid turbovote username. Not required if ``TURBOVOTE_USERNAME``
            env variable set.
        password: str
            A valid turbovote password. Not required if ``TURBOVOTE_PASSWORD``
            env variable set
        subdomain: str
            Your turbovote subdomain (i.e. ``https://MYORG.turbovote.org``). Not
            required if ``TURBOVOTE_SUBDOMAIN`` env variable set.
        uri: str
            Base turbovote API url
    `Returns:`
        class
    """

    def __init__(self, username=None, password=None, subdomain=None):

        self.username = check_env.check('TURBOVOTE_USERNAME', username)
        self.password = check_env.check('TURBOVOTE_PASSWORD', password)
        self.subdomain = check_env.check('TURBOVOTE_SUBDOMAIN', subdomain)
        self.uri = TURBOVOTE_URI

    def _get_token(self):
        # Retrieve a temporary bearer token to access API

        url = self.uri + 'login'

        payload = {'username': self.username,
                   'password': self.password}

        logger.info('Generating token...')
        r = requests.post(url, data=payload)
        logger.debug(r.url)

        r.raise_for_status()

        logger.info('Token generated.')

        return r.json()['id-token']

    def get_users(self):
        """
        Get users.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.uri + f'partners/{self.subdomain}.turbovote.org/users'

        headers = {"Authorization": f"Bearer {self._get_token()}"}

        logger.info('Requesting users table...')
        r = requests.get(url, headers=headers)
        logger.debug(r)
        r.raise_for_status()
        tbl = Table.from_csv_string(r.text)
        logger.info(f'{tbl.num_rows} users retrieved.')

        return tbl



import os
import logging

# Provide shortcuts to importing Parsons submodules

# If you want to be more targeted in your imports, you can set the PARSONS_SKIP_IMPORT_ALL
# environment variable and import classes directly from the Python module where they
# are defined.

if not os.environ.get('PARSONS_SKIP_IMPORT_ALL'):
    from parsons.ngpvan.van import VAN
    from parsons.targetsmart.targetsmart_api import TargetSmartAPI
    from parsons.targetsmart.targetsmart_automation import TargetSmartAutomation
    from parsons.databases.redshift.redshift import Redshift
    from parsons.databases.db_sync import DBSync
    from parsons.aws.s3 import S3
    from parsons.civis.civisclient import CivisClient
    from parsons.etl.table import Table
    from parsons.notifications.gmail import Gmail
    from parsons.google.google_civic import GoogleCivic
    from parsons.google.google_sheets import GoogleSheets
    from parsons.google.google_cloud_storage import GoogleCloudStorage
    from parsons.google.google_bigquery import GoogleBigQuery
    from parsons.phone2action.p2a import Phone2Action
    from parsons.mobilize_america.ma import MobilizeAmerica
    from parsons.facebook_ads.facebook_ads import FacebookAds
    from parsons.notifications.slack import Slack
    from parsons.turbovote.turbovote import TurboVote
    from parsons.sftp.sftp import SFTP
    from parsons.action_kit.action_kit import ActionKit
    from parsons.geocode.census_geocoder import CensusGeocoder
    from parsons.airtable.airtable import Airtable
    from parsons.copper.copper import Copper
    from parsons.crowdtangle.crowdtangle import CrowdTangle
    from parsons.hustle.hustle import Hustle
    from parsons.twilio.twilio import Twilio
    from parsons.salesforce.salesforce import Salesforce
    from parsons.databases.postgres.postgres import Postgres
    from parsons.freshdesk.freshdesk import Freshdesk
    from parsons.bill_com.bill_com import BillCom
    from parsons.newmode.newmode import Newmode
    from parsons.databases.mysql.mysql import MySQL
    from parsons.rockthevote.rtv import RockTheVote
    from parsons.mailchimp.mailchimp import Mailchimp
    from parsons.zoom.zoom import Zoom
    from parsons.action_network.action_network import ActionNetwork

    __all__ = [
        'VAN',
        'TargetSmartAPI',
        'TargetSmartAutomation',
        'Redshift',
        'S3',
        'CivisClient',
        'DBSync',
        'Table',
        'Gmail',
        'GoogleCivic',
        'GoogleCloudStorage',
        'GoogleBigQuery',
        'GoogleSheets',
        'Phone2Action',
        'MobilizeAmerica',
        'FacebookAds',
        'Slack',
        'TurboVote',
        'SFTP',
        'ActionKit',
        'CensusGeocoder',
        'Airtable',
        'Copper',
        'CrowdTangle',
        'Hustle',
        'Twilio',
        'Salesforce',
        'Postgres',
        'Freshdesk',
        'BillCom',
        'Newmode',
        'MySQL',
        'RockTheVote',
        'Mailchimp',
        'Zoom',
        'ActionNetwork',
    ]

# Define the default logging config for Parsons and its submodules. For now the
# logger gets a StreamHandler by default. At some point a NullHandler may be more
# appropriate, so the end user must decide on logging behavior.

logger = logging.getLogger(__name__)
_handler = logging.StreamHandler()
_formatter = logging.Formatter('%(module)s %(levelname)s %(message)s')
_handler.setFormatter(_formatter)
logger.addHandler(_handler)

if os.environ.get('TESTING'):
    # Log less stuff in automated tests
    logger.setLevel('WARNING')
else:
    logger.setLevel('INFO')



"""NGPVAN Events Endpoints"""

from parsons.etl.table import Table
import logging

logger = logging.getLogger(__name__)


class Events(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_events(self, code_ids=None, event_type_ids=None, rep_event_id=None,
                   starting_after=None, starting_before=None, district_field=None,
                   expand_fields=['locations', 'codes', 'shifts', 'roles', 'notes',
                                  'financialProgram', 'ticketCategories',
                                  'onlineForms']):
        """
        Get events.

        `Args:`
            code_ids: str
                Filter by code id.
            event_type_ids: str
                Filter by event_type_ids.
            rep_event_id: str
                Filters to recurring events that are recurrences the passed event id.
            starting_after: str
                Events beginning after ``iso8601`` formatted date.
            starting_before: str
                Events beginning before ``iso8601`` formatted date.
            district_field: str
                Filter by district field.
            expand_fields: list
                A list of fields for which to include data. If a field is omitted,
                ``None`` will be returned for that field. Can be ``locations``, ``codes``,
                ``shifts``,``roles``, ``notes``, ``financialProgram``, ``ticketCategories``,
                ``onlineForms``.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if expand_fields:
            expand_fields = ','.join(expand_fields)

        params = {'codeIds': code_ids,
                  'eventTypeIds': event_type_ids,
                  'inRepetitionWithEventId': rep_event_id,
                  'startingAfter': starting_after,
                  'startingBefore': starting_before,
                  'districtFieldValue': district_field,
                  '$top': 50,
                  '$expand': expand_fields
                  }

        tbl = Table(self.connection.get_request('events', params=params))
        logger.info(f'Found {tbl.num_rows} events.')
        return tbl

    def get_event(self, event_id, expand_fields=['locations', 'codes', 'shifts', 'roles',
                                                 'notes', 'financialProgram', 'ticketCategories',
                                                 'voterRegistrationBatches']):
        """
        Get an event.

        `Args:`
            event_id: int
                The event id.
            expand_fields: list
                A list of fields for which to include data. If a field is omitted,
                ``None`` will be returned for that field. Can be ``locations``,
                ``codes``, ``shifts``, ``roles``, ``notes``, ``financialProgram``,
                ``ticketCategories``, ``voterRegistrationBatches`.`
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if expand_fields:
            expand_fields = ','.join(expand_fields)

        r = self.connection.get_request(f'events/{event_id}', params={'$expand': expand_fields})
        logger.info(f'Found event {event_id}.')
        return r

    def create_event(self, name, short_name, start_date, end_date, event_type_id,
                     roles, shifts=None, description=None, editable=False,
                     publicly_viewable=False, location_ids=None, code_ids=None, notes=None,
                     district_field_value=None, voter_registration_batches=None):
        """
        Create an event

        `Args:`
            name: str
                A name for this event, no longer than 500 characters.
            short_name: str
                A shorter name for this event, no longer than 12 characters.
            start_date: str
                The start date and time for this event.
            end_date: str
                The end date and time for this event that is after ``start_date``
            event_type_id: int
                A valid event type id.
            roles: list
                A list of valid role ids that correspond the with the event type.
            shifts:
                A list of dicts with shifts formatted as:

                .. highlight:: python
                .. code-block:: python

                    [
                        {
                         'name': 'Shift 1',
                         'start_time': '12-31-2018T12:00:00',
                         'end_time': '12-31-2018T13:00:00'
                        }
                        {
                         'name': 'Shift 2',
                         'start_time': '12-31-2018T13:00:00',
                         'end_time': '12-31-2018T14:00:00'
                        }
                    ]

            description: str
                An optional description for this Event, no longer than 500 characters.
            editable: boolean
                If ``True``, prevents modification of this event by any users other than the
                user associated the API key. Setting this to true effectively makes
                the event read-only in the VAN interface.
            publicly_viewable: boolean
                Used by NGP VAN’s website platform to indicate whether this event can be
                viewed publicly.
            location_ids: list
                A list of location_ids where the event is taking place
            code_ids: list
                A list of codes that are applied to this event for organizational purposes. Note
                that at most one source code and any number of tags, may be applied to an event.
            notes: list
                A list of notes
        `Returns:`
            int
              The event code.
        """

        if shifts is None:
            shifts = [{'name': 'Default Shift',
                       'startTime': start_date,
                       'endTime': end_date}]
        else:
            shifts = [{'name': s['name'],
                       'startTime': s['start_time'],
                       'endTime': s['end_time']} for s in shifts]

        event = {'name': name,
                 'shortName': short_name,
                 'description': description,
                 'startDate': start_date,
                 'endDate': end_date,
                 'eventType': {'eventTypeId': event_type_id},
                 'isOnlyEditableByCreatingUser': str(editable).lower(),
                 'isPubliclyViewable': publicly_viewable,
                 'notes': notes,
                 'shifts': shifts,
                 'roles': [{'roleId': r} for r in roles],
                 'districtFieldValue': district_field_value,
                 'voterRegistrationBatches': voter_registration_batches
                 }

        if location_ids:
            event['locations'] = [{'locationId': l} for l in location_ids],

        if code_ids:
            event['codes'] = [{'codeID': c} for c in code_ids]

        r = self.connection.post_request('events', json=event)
        logger.info(f'Event {r} created.')
        return r

    def delete_event(self, event_id):
        """
        Delete an event.

        `Args:`
            event_id: int
                The event id.
        `Returns:`
            ``None``
        """

        r = self.connection.delete_request(f'events/{event_id}')
        logger.info(f'Event {event_id} deleted.')
        return r

    def add_event_shift(self, event_id, shift_name, start_time, end_time):
        """
        Add shifts to an event

        `Args:`
            event_id: int
                The event id.
            shift_name: str
                The name of the shift
            start_time: str
                The start time for the shift (``iso8601`` formatted date).
            end_time: str
                The end time of the shift (``iso8601`` formatted date).
        `Returns:`
            int
              The shift id.
        """

        shift = {'name': shift_name,
                 'startTime': start_time,
                 'endTime': end_time
                 }

        r = self.connection.post_request(f'events/{event_id}/shifts', json=shift)
        logger.info(f'Shift {r} added.')
        return r

    def get_event_types(self):
        """
        Get event types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('events/types'))
        logger.info(f'Found {tbl.num_rows} events.')
        return tbl



from simple_salesforce import Salesforce as _Salesforce
from parsons.utilities import check_env
from parsons.etl import Table
import logging
import json

logger = logging.getLogger(__name__)


class Salesforce:
    """
    `Args:`
        username: str
            The Salesforce username (usually an email address). Not required if
            ``SALESFORCE_USERNAME`` env variable is passed.
        password: str
            The Salesforce password. Not required if ``SALESFORCE_PASSWORD`` env variable is
            passed.
        security_token: str
            The Salesforce security token that can be acquired or reset in
            Settings > My Personal Information > Reset My Security Token.
            Not required if ``SALESFORCE_SECURITY_TOKEN`` env variable is passed.
        test_environment: bool
            If ``True`` the client will connect to a Saleforce sandbox instance. Not required if
            ``SALESFORCE_DOMAIN`` env variable is passed.
    `Returns:`
        Salesforce class
    """

    def __init__(self, username=None, password=None, security_token=None, test_environment=False):

        self.username = check_env.check('SALESFORCE_USERNAME', username)
        self.password = check_env.check('SALESFORCE_PASSWORD', password)
        self.security_token = check_env.check('SALESFORCE_SECURITY_TOKEN', security_token)

        if test_environment:
            self.domain = check_env.check('SALESFORCE_DOMAIN', 'test')
        else:
            self.domain = None

        self._client = None

    def describe_object(self, object):
        """
        `Args:`
            object: str
                The API name of the type of record to describe. Note that custom object names end
                in `__c`
        `Returns:`
            Ordered Dict of all the object's meta data in Salesforce
        """

        return getattr(self.client, object).describe()

    def describe_fields(self, object):
        """
        `Args:`
            object: str
                The API name of the type of record on whose fields you want data. Note that custom
                object names end in `__c`
        `Returns:`
            Dict of all the object's field meta data in Salesforce
        """

        return json.loads(json.dumps(getattr(self.client, object).describe()['fields']))

    def query(self, soql):
        """
        `Args:`
            soql: str
                The desired query in Salesforce SOQL language (SQL with additional limitations).
                For reference, see `Salesforce SOQL documentation<https://developer.salesforce.com/docs/atlas.en-us.soql_sosl.meta/soql_sosl/sforce_api_calls_soql.htm>`_.
        `Returns:`
            list of dicts with Salesforce data
        """ # noqa: E501,E261

        q = Table(self.client.query_all(soql))
        logger.info(f'Found {q.num_rows} results')
        return q

    def insert_record(self, object, data_table):
        """
        Insert new records of the desired object into Salesforce

        `Args:`
            object: str
                The API name of the type of record to insert. Note that custom object names end
                in `__c`
            data_table: obj
                A Parsons Table with data for inserting records. Column names must match object
                field API names, though case and order need not match. Note that custom field
                names end in `__c`.
        `Returns:`
            list of dicts that have the following data:
            * success: boolean
            * created: boolean (if new record is created)
            * id: str (id of record created, if successful)
            * errors: list of dicts (with error details)
        """

        r = getattr(self.client.bulk, object).insert(data_table.to_dicts())
        s = [x for x in r if x.get('success') is True]
        logger.info(
            f'Successfully inserted {len(s)} out of {data_table.num_rows} records to {object}'
        )
        return r

    def update_record(self, object, data_table):
        """
        Update existing records of the desired object in Salesforce

        `Args:`
            object: str
                The API name of the type of record to update. Note that custom object names end
                in `__c`
            data_table: obj
                A Parsons Table with data for updating records. Must contain one column named
                `id`. Column names must match object field API names, though case and order need
                not match. Note that custom field names end in `__c`.
            `Returns:`
                list of dicts that have the following data:
                * success: boolean
                * created: boolean (if new record is created)
                * id: str (id of record altered, if successful)
                * errors: list of dicts (with error details)
        """

        r = getattr(self.client.bulk, object).update(data_table.to_dicts())
        s = [x for x in r if x.get('success') is True]
        logger.info(
            f'Successfully updated {len(s)} out of {data_table.num_rows} records in {object}'
        )
        return r

    def upsert_record(self, object, data_table, id_col):
        """
        Insert new records and update existing ones of the desired object in Salesforce

        `Args:`
            object: str
                The API name of the type of record to upsert. Note that custom object names end
                in `__c`
            data_table: obj
                A Parsons Table with data for upserting records. Column names must match object
                field API names, though case and order need not match. Note that custom field
                names end in `__c`.
            id_col: str
                The column name in `data_table` that stores the record ID. Required even if all
                records are new/inserted.
            `Returns:`
                list of dicts that have the following data:
                * success: boolean
                * created: boolean (if new record is created)
                * id: str (id of record created or altered, if successful)
                * errors: list of dicts (with error details)
        """

        r = getattr(self.client.bulk, object).upsert(data_table.to_dicts(), id_col)
        s = [x for x in r if x.get('success') is True]
        logger.info(
            f'Successfully upserted {len(s)} out of {data_table.num_rows} records to {object}'
        )
        return r

    def delete_record(self, object, id_table, hard_delete=False):
        """
        Delete existing records of the desired object in Salesforce

        `Args:`
            object: str
                The API name of the type of record to delete. Note that custom object names end
                in `__c`
            id_table: obj
                A Parsons Table of record IDs to delete. Note that 'Id' is the default Salesforce
                record ID field name.
            hard_delete: boolean
                If true, will permanently delete record instead of moving it to trash
            `Returns:`
                list of dicts that have the following data:
                * success: boolean
                * created: boolean (if new record is created)
                * id: str (id of record deleted, if successful)
                * errors: list of dicts (with error details)
        """

        if hard_delete:
            r = getattr(self.client.bulk, object).hard_delete(id_table.to_dicts())
        else:
            r = getattr(self.client.bulk, object).delete(id_table.to_dicts())

        s = [x for x in r if x.get('success') is True]
        logger.info(
            f'Successfully deleted {len(s)} out of {id_table.num_rows} records from {object}'
        )
        return r

    @property
    def client(self):
        """
        Get the Salesforce client to use for making all calls. For more information, check the
        `Simple Salesforce Documentation <https://simple-salesforce.readthedocs.io/en/latest/>`_

        `Returns:`
            `simple-salesforce Salesforce object`
        """
        if not self._client:
            # Create a Salesforce client to use to make bulk calls
            self._client = _Salesforce(
                username=self.username,
                password=self.password,
                security_token=self.security_token,
                domain=self.domain
            )

        return self._client
